#!/usr/bin/env cpython
#
# pyLOM - Python Low Order Modeling.
#
# Math operations module - geometry.
#
# Last rev: 27/10/2021

cimport cython
cimport numpy as np

import numpy as np

from .cfuncs       cimport real
from .cfuncs       cimport c_scellCenters, c_snormals, c_seuclidean_d
from .cfuncs       cimport c_dcellCenters, c_dnormals, c_deuclidean_d

from ..utils.cr     import cr
from ..utils.errors import raiseError


@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.float32_t,ndim=2] _scellCenters(float[:,:] xyz, int[:,:] conec):
	'''
	Compute the cell centers given a list 
	of elements (internal function).
	'''
	cdef int nel = conec.shape[0], ndim = xyz.shape[1], ncon = conec.shape[1]
	cdef np.ndarray[np.float32_t,ndim=2] xyz_cen = np.zeros((nel,ndim),dtype = np.float32)
	# Call C function
	c_scellCenters(&xyz_cen[0,0],&xyz[0,0],&conec[0,0],nel,ndim,ncon)
	# Return
	return xyz_cen

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.double_t,ndim=2] _dcellCenters(double[:,:] xyz, int[:,:] conec):
	'''
	Compute the cell centers given a list 
	of elements (internal function).
	'''
	cdef int nel = conec.shape[0], ndim = xyz.shape[1], ncon = conec.shape[1]
	cdef np.ndarray[np.double_t,ndim=2] xyz_cen = np.zeros((nel,ndim),dtype = np.double)
	# Call C function
	c_dcellCenters(&xyz_cen[0,0],&xyz[0,0],&conec[0,0],nel,ndim,ncon)
	# Return
	return xyz_cen

@cr('math.cellCenters')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def cellCenters(real[:,:] xyz, int[:,:] conec):
	r'''
	Compute the cell centers given a list 
	of elements.

	Args:
		xyz (np.ndarray):   node positions
		conec (np.ndarray): connectivity array

	Returns:
		np.ndarray: center positions
	'''
	if real is double:
		return _dcellCenters(xyz,conec)
	else:
		return _scellCenters(xyz,conec)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.float32_t,ndim=2] _snormals(float[:,:] xyz, int[:,:] conec):
	'''
	Compute the cell centers given a list 
	of elements (internal function).
	'''
	cdef int nel = conec.shape[0], ndim = xyz.shape[1], ncon = conec.shape[1]
	cdef np.ndarray[np.float32_t,ndim=2] normals = np.zeros((nel,ndim),dtype = np.float32)
	# Call C function
	c_snormals(&normals[0,0],&xyz[0,0],&conec[0,0],nel,ndim,ncon)
	# Return
	return normals

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.double_t,ndim=2] _dnormals(double[:,:] xyz, int[:,:] conec):
	'''
	Compute the cell centers given a list 
	of elements (internal function).
	'''
	cdef int nel = conec.shape[0], ndim = xyz.shape[1], ncon = conec.shape[1]
	cdef np.ndarray[np.double_t,ndim=2] normals = np.zeros((nel,ndim),dtype = np.double)
	# Call C function
	c_dnormals(&normals[0,0],&xyz[0,0],&conec[0,0],nel,ndim,ncon)
	# Return
	return normals

@cr('math.normals')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def normals(real[:,:] xyz, int[:,:] conec):
	r'''
	Compute the cell normals given a list 
	of elements.

	Args:
		xyz (np.ndarray):   node positions
		conec (np.ndarray): connectivity array

	Returns:
		np.ndarray: cell normals
	'''
	if real is double:
		return _dnormals(xyz,conec)
	else:
		return _snormals(xyz,conec)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.float32_t,ndim=2] _seuclidean_d(float[:,:] X):
	'''
	Compute Euclidean distances between simulations.

	In:
		- X: NxM Data matrix with N points in the mesh for M simulations
	Returns:
		- D: MxM distance matrix 
	'''
	# Initialize
	cdef int n = X.shape[0], m = X.shape[1]
	cdef np.ndarray[np.float32_t,ndim=2] D = np.zeros((m,m),dtype=np.float32)
	# Call C function
	c_seuclidean_d(&D[0,0],&X[0,0],n,m);
	# Return the distance matrix
	return D

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.double_t,ndim=2] _deuclidean_d(double[:,:] X):
	'''
	Compute Euclidean distances between simulations.

	In:
		- X: NxM Data matrix with N points in the mesh for M simulations
	Returns:
		- D: MxM distance matrix 
	'''
	# Initialize
	cdef int n = X.shape[0], m = X.shape[1]
	cdef np.ndarray[np.double_t,ndim=2] D = np.zeros((m,m),dtype=np.double)
	# Call C function
	c_deuclidean_d(&D[0,0],&X[0,0],n,m);
	# Return the distance matrix
	return D

@cr('math.euclidean_d')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def euclidean_d(real[:,:] X):
	r'''
	Compute the Euclidean distances between simulations.

	Args:
		X (np.ndarray): NxM Data matrix with N points in the mesh for M simulations

	Returns:
		np.ndarray: MxM distance matrix 
	'''
	if real is double:
		return _deuclidean_d(X)
	else:
		return _seuclidean_d(X)


# Example of PINN on Burgers equation.
# The PDE is:
#   u_t + u * u_x - 0.01 * u_xx = 0, x in [-1, 1], t in [0, 1]
# with initial condition:
#   u(0, x) = -sin(pi * x), x in [-1, 1]
# and boundary conditions:
#   u(t, -1) = u(t, 1) = 0, t in [0, 1]


import sys, os, numpy as np
import torch
import pyLOM, pyLOM.NN


DATAFILE  = sys.argv[1]
VARIABLES = eval(sys.argv[2])
OUTDIR    = sys.argv[3]

device  = 'cpu'
RESUDIR = os.path.join(OUTDIR,f'PINN_{DATAFILE}')
pyLOM.NN.create_results_folder(RESUDIR,verbose=False)

# Define the domain and the amount of points to sample
POINTS_ON_X = 256
POINTS_ON_T = 100
num_train_simulations = 4000
t = np.linspace(0, 1, POINTS_ON_T)
x = np.linspace(-1, 1, POINTS_ON_X)
T, X = np.meshgrid(t, x)

idx = np.random.choice(X.flatten().shape[0], num_train_simulations, replace=False)
TX = np.concatenate([T.reshape(-1, 1), X.reshape(-1, 1)], axis=1)
TX = torch.tensor(TX).float()
train_TX = TX[idx]

# Define the boundary conditions
class InitialCondition(pyLOM.NN.BoundaryCondition):

    def loss(self, pred):
        x = self.points[:, 1].reshape(-1, 1)
        initial_cond_pred = pred
        # sin is positive here because the initial condition is u(0, x) = -sin(pi * x)
        ic_loss = (initial_cond_pred + torch.sin(torch.pi * x).to(device)) ** 2
        return ic_loss.mean()


class XBoudaryCondition(pyLOM.NN.BoundaryCondition):

    def loss(self, pred):
        # as u on the boundary is 0, we can just return the mean of the prediction
        return pred.pow(2).mean()


initial_points = torch.tensor(x).reshape(-1, 1)
initial_bc = InitialCondition(
    torch.cat([torch.full_like(initial_points, 0), initial_points], dim=-1).float(),
)

boundary_points = torch.tensor(t).reshape(-1, 1)
boundary_bc = XBoudaryCondition(
    torch.cat(
        [torch.cat([boundary_points, torch.full_like(boundary_points, -1)], dim=-1),
        torch.cat([boundary_points, torch.full_like(boundary_points, 1)], dim=-1),]
    ).float()
)

train_dataset = torch.utils.data.TensorDataset(train_TX)
test_dataset = torch.utils.data.TensorDataset(TX)

# Define the neural network
input_dim = TX.shape[1]
output_dim = 1 # u(t, x)

# The pinn needs a neural network, which can be any pytorch model that implements nn.Module
net = pyLOM.NN.MLP(
    input_size=input_dim,
    output_size=output_dim,
    hidden_size=40,
    n_layers=3,
    activation=torch.nn.functional.tanh, # With relu the model struggles to converge
)

burgers_pinn = pyLOM.NN.BurgersPINN(
    viscosity=0.01,
    neural_net=net,
    device=device,
)
pyLOM.pprint(0, burgers_pinn)

training_params = {
    'optimizer_class': torch.optim.Adam,
    'optimizer_params': {'lr': 1e-3},
    'lr_scheduler_class': torch.optim.lr_scheduler.StepLR,
    'lr_scheduler_params': {'step_size': 1000, 'gamma': 0.99},
    'epochs': 100,
    'update_logs_steps': 100,
    'boundary_conditions': [initial_bc, boundary_bc],
}

# Create the pipeline and train the model with Adam
pipeline_adam = pyLOM.NN.Pipeline(
    model=burgers_pinn,
    train_dataset=train_dataset,
    test_dataset=test_dataset,
    training_params=training_params
)
model_logs = pipeline_adam.run()



# Train the model with L-BFGS to improve the results
lbfgs_params = {
    'lr': 0.01,
    'max_iter': 12000,
    'max_eval': 10000,
    'history_size': 200,
    'tolerance_grad': 1e-12,
    'tolerance_change': 0.5 * np.finfo(float).eps,
    'line_search_fn': 'strong_wolfe'
}
training_params = {
    'optimizer_class': torch.optim.LBFGS,
    'optimizer_params': lbfgs_params,
    'loaded_logs': model_logs,
    'epochs': 1,
    'boundary_conditions': [initial_bc, boundary_bc],
}

pipeline_lbfgs = pyLOM.NN.Pipeline(
    model=burgers_pinn,
    train_dataset=train_dataset,
    test_dataset=test_dataset,
    training_params=training_params
)
model_logs = pipeline_lbfgs.run()
pyLOM.pprint(0, f"Residual: {model_logs['test_loss'][-1]}")

burgers_pinn.save(RESUDIR + '/burgers_pinn.pt')
pyLOM.pprint(0, "Model saved")
burgers_pinn_loaded = pyLOM.NN.BurgersPINN.load(RESUDIR + '/burgers_pinn.pt', device=device)

pyLOM.pprint(0, "Model loaded")

u = burgers_pinn.predict(test_dataset).reshape(POINTS_ON_X, POINTS_ON_T)

## Testsuite output
pyLOM.pprint(0,'TSUITE u     =',u.min(),u.max(),u.mean())

pyLOM.cr_info()
pyLOM.pprint(0,'End of output')


#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# Math operations Module
#
# Last rev: 27/10/2021

# Vector matrix routines
from .maths      import transpose, vector_sum, vector_norm, vector_mean, matmul, matmulp, vecmat, argsort, eigen, polar, cholesky, vandermonde, conj, diag, inv, flip, vandermondeTime
# Averaging routines
from .averaging  import temporal_mean, subtract_mean
# Truncation routines
from .truncation import compute_truncation_residual, energy
# Statistics routines
from .stats      import RMSE, MAE, r2
# SVD routines
from .svd        import qr, svd, tsqr, randomized_qr, init_qr_streaming, update_qr_streaming, tsqr_svd, randomized_svd
# FFT routines
from .fft        import hammwin, fft
# Geometry and mesh routines
from .geometric  import cellCenters, normals, euclidean_d
# Regression routines
from .regression import least_squares, ridge_regresion


del maths, averaging, truncation, stats, geometric, regression



#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# Last rev: 09/07/2021

__version__ = '3.0.0'

# Import essential tools
from .                import inp_out as io, vmmath as math
from .dataset         import Dataset
from .partition_table import PartitionTable
from .mesh            import Mesh

# Import utilities
from .utils.cr     import cr, cr_nvtx, cr_start, cr_stop, cr_reset, cr_info
from .utils.nvtxp  import nvtxp
from .utils.parall import pprint
from .utils.gpu    import gpu_device
from .utils.plots  import show_plots, close_plots

# Import Low Order Models
from . import POD, DMD, SPOD, MANIFOLD, GPOD

# Import AI Models
# Leaving this commented as the NN module overloads the memory when loaded
# to load the NN module use `import pyLOM.NN` instead
# Show a warning in case the modules cannot be loaded
#try:
#    from . import NN
#except:
#	from .utils.errors import raiseWarning
#	raiseWarning('Import - Cannot load NN! Ensure that the correct dependencies are installed',allranks=False)
#	del raiseWarning


del dataset, partition_table, mesh



#!/usr/bin/env python
#
# POD analysis.
#
# Last revision: 14/06/2024
import numpy as np
import matplotlib.pyplot as plt
import pyLOM


def generate_X():
    x = np.linspace(0, 1, 100)
    y = np.linspace(0, 1, 200)
    z = np.linspace(0, 1, 300)
    XX, YY, ZZ = np.meshgrid(x, y, z)
    F = np.sin(2*np.pi*(XX + ZZ)) + np.cos(2*np.pi*(2*YY + ZZ))
    return F

# Generate synthetic database
X = generate_X()
print('Shape of synthetic database:',X.shape) # (200, 100, 300)

#########################################################################################################
## However, pyLOM takes a database of the shape (spatial_dims,time_dims)
## thus we need to reshape this matrix accordingly.
#########################################################################################################
X_rsh = X.reshape([-1, X.shape[-1]])
print('Shape of reshaped database:',X_rsh.shape) # (200*100 = 20000, 300)


#########################################################################################################
## POD using the whole database.
#########################################################################################################
PSI, S, V = pyLOM.POD.run(X_rsh, remove_mean=False)
PSI_trunc, S_trunc, V_trunc = pyLOM.POD.truncate(PSI,S,V,r=1e-6)
X_rsh_rec = pyLOM.POD.reconstruct(PSI_trunc, S_trunc, V_trunc)
print('Shape of reconstructed database:',X_rsh_rec.shape) # (200*100 = 20000, 300)
print("RMSE for whole database =", pyLOM.math.RMSE(X_rsh_rec,X_rsh))

fig, axs = plt.subplots(1, 2, figsize=(8, 4))
axs[0].pcolor(X_rsh[:,0].reshape((200,100)))
axs[0].set_title("Original matrix")
axs[1].pcolor(X_rsh_rec[:,0].reshape((200,100)))
axs[1].set_title("Reconstructed matrix")
plt.show()


########################################################################################################
# POD using just the first "snapshot" (as if we are doing data compression).
########################################################################################################
X_rsh_2 = X_rsh[:,:1].copy()
PSI, S, V = pyLOM.POD.run(X_rsh_2, remove_mean=False)
PSI_trunc, S_trunc, V_trunc = pyLOM.POD.truncate(PSI,S,V,r=1e-6)
X_rsh_rec = pyLOM.POD.reconstruct(PSI_trunc, S_trunc, V_trunc)
print('Shape of reconstructed database:',X_rsh_rec.shape) # (200*100 = 20000, 0)
print("RMSE for whole database =", pyLOM.math.RMSE(X_rsh_rec,X_rsh_2))

fig, axs = plt.subplots(1, 2, figsize=(8, 4))
axs[0].pcolor(X_rsh_2[:,0].reshape((200,100)))
axs[0].set_title("Original matrix")
axs[1].pcolor(X_rsh_rec[:,0].reshape((200,100)))
axs[1].set_title("Reconstructed matrix")
plt.show()

## Print timings
pyLOM.cr_info()


#!/usr/bin/env python
#
# pyLOM, utils.
#
# Parallel routines
#
# Last rev: 14/02/2025
from __future__ import print_function, division

import mpi4py, numpy as np
mpi4py.rc.recv_mprobe = False
from mpi4py import MPI

import numpy as np

MPI_COMM = MPI.COMM_WORLD
MPI_RANK = MPI_COMM.Get_rank()
MPI_SIZE = MPI_COMM.Get_size()

mpi_send = MPI_COMM.send
mpi_recv = MPI_COMM.recv


def worksplit(istart,iend,whoAmI,nWorkers=MPI_SIZE):
	'''
	Divide the work between the processors
	'''
	istart_l, iend_l = istart, iend
	irange = iend - istart
	if (nWorkers < irange):
		# We split normally among processes assuming no remainder
		rangePerProcess = int(np.floor(irange/nWorkers))
		istart_l = istart   + whoAmI*rangePerProcess
		iend_l   = istart_l + rangePerProcess
		# Handle the remainder
		remainder = irange - rangePerProcess*nWorkers
		if remainder > whoAmI:
			istart_l += whoAmI
			iend_l   += whoAmI+1;
		else:
			istart_l += remainder
			iend_l   += remainder
	else:
		# Each process will forcefully conduct one instant.
		istart_l = whoAmI   if whoAmI < iend else iend
		iend_l   = whoAmI+1 if whoAmI < iend else iend

	return istart_l, iend_l


def writesplit(npoints,write_master):
	'''
	Divide the write array between the processors
	'''
	rstart = 1 if not write_master else 0
	istart, iend = 0, 0 
	# Select in which order the processors will write
	if MPI_RANK == rstart:
		# send to next where to start writing
		istart, iend = 0, npoints
		mpi_send(iend,dest=MPI_RANK+1)
	elif MPI_RANK == MPI_SIZE-1:
		# recive from the previous where to start writing
		istart = mpi_recv(source=MPI_RANK-1) 
		iend   = istart + npoints
	else:
		# recive from the previous where to start writing
		istart = mpi_recv(source=MPI_RANK-1) 
		iend   = istart + npoints
		# send to next where to start writing
		mpi_send(iend,dest=MPI_RANK+1) 
	return istart, iend

def split(array,root=0):
	'''
	Split an array among the processors
	'''
	return np.vsplit(array,[worksplit(0,array.shape[0],i)[1] for i in range(MPI_SIZE-1)]) if MPI_RANK==root else None

def is_rank_or_serial(root=0):
	'''
	Return whether the rank is active or True
	in case of a serial run
	'''
	return MPI_RANK == root or MPI_SIZE == 1

def pprint(rank,*args,**kwargs):
	'''
	Print alternative for parallel codes. It works as
	python's print with the rank variable, which can 
	be negative for everyone to print or equal to the
	rank that should print.
	'''
	if MPI_SIZE == 1:
		print(*args,**kwargs)
	elif rank < 0:
		print('Rank %d:'%MPI_RANK,*args,**kwargs)
	elif rank == MPI_RANK:
		print('Rank %d:'%rank,*args,**kwargs)



#!/usr/bin/env python
#
# Example of GPR.
#
# Last revision: 19/02/2025
from __future__ import print_function, division

import sys, os, json, numpy as np
import pyLOM, pyLOM.GPR

## Parameters
DATAFILE  = sys.argv[1]
VARIABLES = eval(sys.argv[2])
OUTDIR    = sys.argv[3]
PARAMS    = json.loads(str(sys.argv[4]).replace("'",'"'))

## Data loading
d = pyLOM.Dataset.load(DATAFILE)

# Data selection
selected_snapshot = 5
pressure_snapshot = d.X(*VARIABLES)[:, selected_snapshot]
coordinates = d.xyz

np.random.seed(42)
masked_pressure = pressure_snapshot*(np.random.rand(len(pressure_snapshot)) < 0.65)
nonzero_indices = np.nonzero(masked_pressure)[0]
filtered_coordinates = coordinates[nonzero_indices]

x_train = filtered_coordinates
y_train = masked_pressure[nonzero_indices]
x_pred = coordinates

# GPR
gpr = pyLOM.GPR.SF_GPR(input_dim=2)
print(gpr.kernel.available_kernels)  # Dict of available kernels
print(
    gpr.kernel.get_kernel_parameters("RBF")
)  # Dict of available parameters for the selecter kernel
kernel = gpr.kernel.Matern32(variance=0.1, lengthscale=1.0, ARD=True)
gpr.fit(x_train, y_train, kernel)
gpr.display_model()
result = gpr.predict(x_pred)
mean = result["mean"]
std = result["std"]

rmse = pyLOM.math.RMSE(result["mean"].flatten(),pressure_snapshot)

## Testsuite output
pyLOM.pprint(0,'TSUITE RMSE      = %e'%rmse)
pyLOM.pprint(0,'TSUITE GPR MEAN  =',mean.min(),mean.max(),mean.mean())
pyLOM.pprint(0,'TSUITE GPR STD   =',std.min(),std.max(),std.mean())

## Show and print timings
pyLOM.cr_info()
pyLOM.pprint(0,'End of output')


#!/usr/bin/env python
#
# PYLOM Testsuite GPOD.
#
# Last revision: 26/02/2025
from __future__ import print_function, division
import mpi4py
import pyLOM.GPOD

mpi4py.rc.recv_mprobe = False

import sys, json, numpy as np
import pyLOM

## Parameters
DATAFILE  = sys.argv[1]
VARIABLES = eval(sys.argv[2])
OUTDIR    = sys.argv[3]
PARAMS    = json.loads(str(sys.argv[4]).replace("'",'"').lower())

## Data loading
m = pyLOM.Mesh.load(DATAFILE)
d = pyLOM.Dataset.load(DATAFILE, ptable=m.partition_table)
X = d.X(*VARIABLES)
t = d.get_variable('time')


##########################################
# Reconstruct the Gappy Snapshot Matrix
##########################################
# Generate random mask
np.random.seed(5)
random_mask = np.random.choice(
    [0, 1], size=X.shape, p=[PARAMS['missing_percentage'], 1 - PARAMS['missing_percentage']]
)
# Create incomplete snapshot matrix
incomplete_snapshot = X * random_mask


# Database reconstruction
num_iter = 20

# GappyPOD model setup
gappy_model_recons = pyLOM.GPOD.GappyPOD(**PARAMS['gpod'])

# Reconstruct database
X_recons, eig_spec_iter, c_e = gappy_model_recons.reconstruct_full_set(
    incomplete_snapshot,
    PARAMS['num_iter'],
)

# Compute metrics
mae_recons  = pyLOM.math.MAE(X, X_recons)
rmse_recons = pyLOM.math.RMSE(X, X_recons, relative=False)
r2_recons   = pyLOM.math.r2(X, X_recons)


## Testsuite output
pyLOM.pprint(0,'TSUITE VELOX =',X_recons.min(),X_recons.max(),X_recons.mean())
pyLOM.pprint(0,'TSUITE MAE   = %e'%mae_recons)
pyLOM.pprint(0,'TSUITE RMSE  = %e'%rmse_recons)
pyLOM.pprint(0,'TSUITE R2    = %e'%r2_recons)


## Show and print timings
pyLOM.cr_info()
pyLOM.pprint(0,'End of output')



#!/usr/bin/env cpython
#
# pyLOM - Python Low Order Modeling.
#
# Math operations module - averaging.
#
# Last rev: 27/10/2021
from __future__ import print_function, division

import numpy as np

from ..utils.gpu import cp
from ..utils     import cr_nvtx as cr


## Python functions
@cr('math.temporal_mean')
def temporal_mean(X:np.ndarray) -> np.ndarray:
	r'''
	Temporal mean of matrix X(m,n) where m is the spatial coordinates
	and n is the number of snapshots.

	Args:
		X (numpy.ndarray): Snapshot matrix (m,n).

	Returns:
		numpy.ndarray: Averaged snapshot matrix (m,).
	'''
	p = cp if type(X) is cp.ndarray else np
	return p.mean(X,axis=1)

@cr('math.subtract_mean')
def subtract_mean(X:np.ndarray,X_mean:np.ndarray) -> np.ndarray:
	r'''
	Computes out(m,n) = X(m,n) - X_mean(m) where m is the spatial coordinates
	and n is the number of snapshots.

	Args:
		X (numpy.ndarray): Snapshot matrix (m,n).
		X_mean (numpy.ndarray): Averaged snapshot matrix (m,)

	Returns:
		numpy.ndarray: Snapshot matrix without the average(m,n).
	'''
	p = cp if type(X) is cp.ndarray else np
	return X - p.tile(X_mean,(X.shape[1],1)).T


#!/usr/bin/env python
#
# Example of 3D-VAE.
#
# Last revision: 24/09/2024
from __future__ import print_function, division
 
import mpi4py
mpi4py.rc.recv_mprobe = False

import os, numpy as np
import pyLOM, pyLOM.NN


## Set device
device = pyLOM.NN.select_device("cpu") # Force CPU for this example, if left in blank it will automatically select the device


## Specify autoencoder parameters
ptrain      = 0.8
pvali       = 0.2
batch_size  = 4
nepochs     = 10
nlayers     = 4
channels    = 48
lat_dim     = 10
beta        = 1e-04
beta_wmup   = 500
beta_start  = 30
kernel_size = 4
nlinear     = 256
padding     = 1
activations = [pyLOM.NN.relu(), pyLOM.NN.relu(), pyLOM.NN.relu(), pyLOM.NN.relu(), pyLOM.NN.relu(), pyLOM.NN.relu()]
batch_norm  = False
vae         = True


## Load dataset and set up the results output
BASEDIR  = './DATA/'
CASESTR  = 'CHANNEL'
DSETDIR  = os.path.join(BASEDIR,f'{CASESTR}.h5')
VARIABLE = 'VELOX'
RESUDIR  = 'vae_beta_%.2e_ld_%i' % (beta, lat_dim)
pyLOM.NN.create_results_folder(RESUDIR)


## Load the dataset
m    = pyLOM.Mesh.load(DSETDIR) # Mesh size (100 x 40 x 64)
d    = pyLOM.Dataset.load(DSETDIR,ptable=m.partition_table)
u    = d[VARIABLE] # vars ['VELOC'] : u
um   = pyLOM.math.temporal_mean(u)
u_x  = pyLOM.math.subtract_mean(u, um)
time = d.get_variable('time') # 120 instants
pyLOM.pprint(0,"Variables: ", d.varnames)
pyLOM.pprint(0,"Information about the variable: ", d.info(VARIABLE))
pyLOM.pprint(0,"Number of points: ", len(d))
pyLOM.pprint(0,"Instants: ", time.shape[0])


## Mesh Size
n0x = len(np.unique(np.round(d.xyz[:,0],5)))
n0y = len(np.unique(np.round(d.xyz[:,1],5))) 
n0z = len(np.unique(np.round(d.xyz[:,2],5))) 
nx, ny, nz = 64, 64, 64


## Create the torch dataset
td = pyLOM.NN.Dataset((u_x,), (n0x, n0y, n0z))
td.crop(nx, ny, nz)


## Set and train the variational autoencoder
betasch    = pyLOM.NN.betaLinearScheduler(0., beta, beta_start, beta_wmup)
encoder    = pyLOM.NN.Encoder3D(nlayers, lat_dim, nx, ny, nz, td.num_channels, channels, kernel_size, padding, activations, nlinear, batch_norm=batch_norm, stride = 2, dropout = 0, vae = vae)
decoder    = pyLOM.NN.Decoder3D(nlayers, lat_dim, nx, ny, nz, td.num_channels, channels, kernel_size, padding, activations, nlinear, batch_norm=batch_norm)
model      = pyLOM.NN.VariationalAutoencoder(lat_dim, (nx, ny, nz), td.num_channels, encoder, decoder, device=device)
early_stop = pyLOM.NN.EarlyStopper(patience=15, min_delta=0.05)

pipeline = pyLOM.NN.Pipeline(
    train_dataset   = td,
    test_dataset    = td,
    model           = model,
    training_params = {
        "batch_size": 4,
        "epochs": 500,
        "lr": 1e-4,
        "betasch": betasch,
        "BASEDIR": RESUDIR
    },
)
pipeline.run()


## Reconstruct dataset and compute accuracy
rec = model.reconstruct(td)
rd  = pyLOM.NN.Dataset((rec,), (nx, ny, nz))
rd.pad(n0x, n0y, n0z)
td.pad(n0x, n0y, n0z)
d.add_field('urec',1,rd[:,0,:,:].numpy().reshape((len(time),n0x*n0y*n0z)).T)
d.add_field('utra',1,td[:,0,:,:].numpy().reshape((len(time),n0x*n0y*n0z)).T)
pyLOM.io.pv_writer(m,d,'reco',basedir=RESUDIR,instants=np.arange(time.shape[0],dtype=np.int32),times=time,vars=['urec','VELOX','utra'],fmt='vtkh5')
pyLOM.NN.plotSnapshot(m,d,vars=['urec'],instant=0,component=0,cmap='jet')
pyLOM.NN.plotSnapshot(m,d,vars=['utra'],instant=0,component=0,cmap='jet')

pyLOM.cr_info()


#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
#
# Last rev: 07/02/2025
from __future__ import print_function, division

import numpy as np

from ..utils.gpu import cp


def delete_snapshot(X, snap, axis=1):
	"""
	Remove snapshot from training data.
	"""
	p = cp if type(X) is cp.ndarray else np
	return p.delete(X, snap, axis=axis)


def set_random_elements_to_zero(vector, percentage):
	"""
	Randomly sets a specified percentage of elements in a vector to zero.

	Args:
		vector (np.ndarray): Input vector.
		percentage (float): Percentage of elements to set to zero.

	Returns:
		np.ndarray: Modified vector with zeros.
	"""
	p = cp if type(vector) is cp.ndarray else np
	modified_vector = vector.copy()
	num_zeros = int(len(vector) * percentage / 100)
	indices = p.random.choice(len(vector), num_zeros, replace=False)
	modified_vector[indices] = 0
	return modified_vector



#!/usr/bin/env python
#
# pyLOM, dataset.
#
# Dataset class, reader and reduction routines.
#
# Last rev: 30/07/2021
from __future__ import print_function, division

import numpy as np

from .utils import cr, mem, MPI_SIZE, worksplit, mpi_gather


class PartitionTable(object):
	'''
	The partition table class contains information on the 
	partition used for the given dataset or  it can generate
	a new partition
	'''
	def __init__(self,nparts,ids,elements,points,has_master=False):
		'''
		Class constructor
		'''
		self._nparts   = nparts
		self._ids      = ids
		self._elements = elements
		self._master   = has_master if MPI_SIZE > 1 else False
		self._points   = points
		self._inods    = None

	def __str__(self):
		out  = 'Partition Table:\nnumber of partitions: %d\n' % self.n_partitions
		out += '\tIds  |  Elements  |  Points  \n'
		for ipart in range(self.n_partitions):
			out += '\t %03d |    %04d    |    %04d \n' %(self.Ids[ipart],self.Elements[ipart],self.Points[ipart])
		return out

	@cr('PartTable.pbounds')
	def partition_bounds(self,rank,ndim=1,points=True):
		'''
		Compute the partition bounds for a given rank
		'''
		if self._master and rank == 0 and not MPI_SIZE == 1: 
			return 0, 1
		offst    = 1 if not self._master else 0
		mask_idx = self.Ids < rank + offst
		this_idx = self.Ids == rank + offst
		table    = self.Points if points else self.Elements
		istart   = np.sum(table[mask_idx])*ndim
		iend     = istart + table[this_idx][0]*ndim
		return istart, iend

	@cr('PartTable.set_ppoints')
	def create_partition_points(self,conec):
		'''
		Find which nodes this partition has
		'''
		self._inods = np.unique(conec.flatten())
		self.update_points(self._inods.shape[0])

	@cr('PartTable.ppoints')
	def partition_points(self,npoints,ndim=1):
		'''
		Compute the points to be read for this partition
		'''
		mynods  = np.array([],np.int32)
		# Deal with multiple dimensions
		for idim in range(ndim):
			mynods = np.hstack((mynods,self._inods+idim*npoints))
		return mynods		

	@cr('PartTable.reorder')
	def reorder_points(self,xyz,conectivity):
		'''
		Reorder the points array so that in matches with
		the partition table, in serial algorithm.
		'''
		xyz_new = np.zeros_like(xyz)
		# Loop all the partitions
		for ipart in range(self.n_partitions):
			mynods = self.partition_points(ipart,conectivity)
			# Rearrange the node vector
			nstart  = np.cumsum(self.Points[:ipart])
			nend    = self.Points[ipart] + nstart
			xyz_new[nstart:nend] = xyz[mynods]
		# Return
		return xyz_new

	def update_points(self,npoints_new):
		'''
		Update the number of points on the table
		'''
		p = mpi_gather(npoints_new,all=True)
		self._points = p if isinstance(p,np.ndarray) else np.array([p],np.int32)

	def check_split(self):
		'''
		See if a table has the same number of subdomains
		than the number of mpi ranks
		'''
		# Deal with master and serial
		offst = 1 if self._master and not MPI_SIZE == 1 else 0
		return self._nparts + offst == MPI_SIZE

	@classmethod
	@cr('PartTable.new')
	def new(cls,nparts,nelems,npoints,has_master=False):
		'''
		Create a new partition table, in serial algorithm.
		'''
		ids      = np.zeros((nparts,),np.int32)
		points   = np.zeros((nparts,),np.int32)
		elements = np.zeros((nparts,),np.int32)
		# For all the partitions do
		for ipart in range(nparts):
			ids[ipart] = ipart + 1
			# Split the number of elements
			istart, iend = worksplit(0,nelems,ipart,nWorkers=nparts)
			# How many elements do I have
			elements[ipart] = iend - istart
			# How many nodes do I have
			istart, iend  = worksplit(0,npoints,ipart,nWorkers=nparts)
			points[ipart] = iend - istart
		return cls(nparts,ids,elements,points,has_master=has_master)

	@classmethod
	@cr('PartTable.from_pyQvarsi')
	def from_pyQvarsi(cls,ptable,porder=1,has_master=False):
		'''
		Create a partition table from a partition table coming
		from pyQvarsi
		'''
		nparts   = ptable.n_partitions
		ids      = np.arange(1,nparts+1,dtype=np.int32)
		points   = ptable.Points
		elements = ptable.Elements*porder**3
		return cls(nparts,ids,elements,points,has_master=has_master)		

	@property
	def n_partitions(self):
		return self._nparts
	@property
	def Ids(self):
		return self._ids
	@property
	def Elements(self):
		return self._elements
	@property
	def Points(self):
		return self._points
	@property
	def has_master(self):
		return self._master
	@property
	def nodes(self):
		return self._inods
	@nodes.setter
	def nodes(self,inods):
		self._inods = inods



#!/bin/env python
# 
# Test pyLOM multi-gpu using cupy
# Efficiency test, all GPUs have the same size matrix
#
#  Matrix size guidelines depending on the machine tested:
#
#### A node of MareNostrum 5 (4 NVIDIA H100 GPUs per node with 64 Gb of VRAM each) can handle the following sizes:
######## M = 8e6, N = 500
######## M = 4e6, N = 1000
######## M = 2e6, N = 2000
#
#### A single GPU of Juno 3 (48 Gb of VRAM) can handle the following sizes:
######## M = 8e6, N = 250
######## M = 4e6, N = 500
######## M = 2e6, N = 1000
#
#### Add any other tested machine ####
# Last revision: 26/02/2025
from __future__ import print_function, division

import numpy as np, cupy as cp
import pyLOM
import sys

from pyLOM.utils.mpi import MPI_SIZE

pyLOM.gpu_device(gpu_per_node=4) # MN5 has 4 GPU per node

# Define the global matrix size according to the capabilities of your machine
M = int(np.double(sys.argv[1]))
N = int(sys.argv[2])

# Generate a large enough random matrix on the GPU
Ai = cp.random.rand(M,N).astype(cp.float32)

# Run pyLOM TSQR-SVD algorithm
Ui, S, V = pyLOM.math.tsqr_svd(Ai)

# Print results
pyLOM.pprint(0,f'RUNNING WITH {MPI_SIZE} GPUs')
pyLOM.cr_info()



#!/usr/bin/env python
#
# Conversor of MAT file to HDF5 file.
#
# Last revision: 03/08/2021
from __future__ import print_function, division

import numpy as np, h5py
import pyLOM


## Parameters
MATFILE = '../DATA/Tensor_re280.mat'
OUTFILE = './DATA/TENSOR280.h5'
DT      = 1.
DIMSX   = 0., 10.
DIMSY   = -2., 2.
DIMSZ   = 0., 1.67


## Load MAT file
f   = h5py.File(MATFILE,'r')
mat = { k:np.array(v) for k,v in f.items() }
f.close()

tensor = np.ascontiguousarray(np.transpose(mat['Tensor'],(4,3,2,1,0)))
[ndim,nx,ny,nz,nt] = tensor.shape


## Build mesh information dictionary
mesh = pyLOM.Mesh.new_struct3D(nx+1,ny+1,nz+1,None,None,None,DIMSX,DIMSY,DIMSZ) 
print(mesh)


## Create partition table
ptable = pyLOM.PartitionTable.new(1,mesh.ncells,mesh.npoints)
mesh.partition_table = ptable
print(ptable)


## Build time instants
time = DT*np.arange(nt) + DT
time = time[280:400]
nt   = time.shape[0]


## Obtain 3D velocity field
npoints = nx*ny*nz
VELOX = tensor[0,:,:,:,280:400].reshape((nx*ny*nz,nt),order='C')
VELOY = tensor[1,:,:,:,280:400].reshape((nx*ny*nz,nt),order='C')
VELOZ = tensor[2,:,:,:,280:400].reshape((nx*ny*nz,nt),order='C')


## Create dataset for pyLOM
nsample = 10
d = pyLOM.Dataset(xyz=mesh.xyzc, ptable=ptable, order=mesh.cellOrder, point=False,
	# Add the time as the only variable
	vars  = {'time':{'idim':0,'value':time[::nsample]}},
	# Now add all the arrays to be stored in the dataset
	# It is important to convert them as C contiguous arrays
	VELOX = {'ndim':1,'value':VELOX[:,::nsample].copy()},
)
print(d)

mesh.save(OUTFILE) # Store the mesh
d.save(OUTFILE)    # Store dataset


pyLOM.cr_info()


#!/usr/bin/env python
#
# PYLOM Testsuite
# Run DMD on the cylinder dataset
#
# Last revision: 27/01/2023
from __future__ import print_function, division

import sys, os, numpy as np
import pyLOM


## Parameters
DATAFILE  = sys.argv[1]
VARIABLES = eval(sys.argv[2])
OUTDIR    = sys.argv[3]


## Data loading
m  = pyLOM.Mesh.load(DATAFILE)
d  = pyLOM.Dataset.load(DATAFILE,ptable=m.partition_table)
X = d.X(*VARIABLES)
t  = d.get_variable('time')
dt = t[1] - t[0]


## Run DMD
Y = X[:,:100].copy() # Grab the first 100 and run DMD
muReal,muImag,Phi,bJov = pyLOM.DMD.run(Y,1e-6,remove_mean=False)
# Compute frequency and damping ratio of the modes
delta, omega = pyLOM.DMD.frequency_damping(muReal,muImag,dt)
os.makedirs(OUTDIR,exist_ok=True)
pyLOM.DMD.save(f'{OUTDIR}/results.h5',muReal,muImag,Phi,bJov,d.partition_table,nvars=len(VARIABLES),pointData=d.point)
# Reconstruction according to Jovanovic 2014
# Last 50 snapshots are predicted
X_DMD = pyLOM.DMD.reconstruction_jovanovic(Phi,muReal,muImag,t,bJov)
rmse  = pyLOM.math.RMSE(X_DMD.copy(),X.copy())


## Testsuite output
M_real = pyLOM.DMD.extract_modes(Phi,1,len(d),real=True,modes=[1,4,6,2,5,3])
M_imag = pyLOM.DMD.extract_modes(Phi,1,len(d),real=False,modes=[1,4,6,2,5,3])
pyLOM.pprint(0,'TSUITE RMSE   = %e'%rmse)
pyLOM.pprint(0,'TSUITE muReal =',muReal.min(),muReal.max(),muReal.mean())
pyLOM.pprint(0,'TSUITE muImag =',muImag.min(),muImag.max(),muImag.mean())
pyLOM.pprint(0,'TSUITE Phi    =',Phi.min(),Phi.max(),Phi.mean())
pyLOM.pprint(0,'TSUITE bJov   =',bJov.min(),bJov.max(),bJov.mean())
pyLOM.pprint(0,'TSUITE delta  =',delta.min(),delta.max(),delta.mean())
pyLOM.pprint(0,'TSUITE omega  =',omega.min(),omega.max(),omega.mean())
pyLOM.pprint(0,'TSUITE Y      =',Y.min(),Y.max(),Y.mean())
pyLOM.pprint(0,'TSUITE X_DMD  =',X_DMD.min(),X_DMD.max(),X_DMD.mean())
pyLOM.pprint(0,'TSUITE M_real =',M_real.min(),M_real.max(),M_real.mean())
pyLOM.pprint(0,'TSUITE M_imag =',M_imag.min(),M_imag.max(),M_imag.mean())


## DMD plots
if pyLOM.utils.is_rank_or_serial(0):
	fig,_ = pyLOM.DMD.ritzSpectrum(muReal,muImag)               # Ritz Spectrum
	fig.savefig(f'{OUTDIR}/ritzSpectrum.png',dpi=300)
	fig,_ = pyLOM.DMD.amplitudeFrequency(omega,bJov,norm=False) # Amplitude vs frequency
	fig.savefig(f'{OUTDIR}/amplitudeFrequency.png',dpi=300)
	fig,_ = pyLOM.DMD.dampingFrequency(omega,delta)             # Damping ratio vs frequency
	fig.savefig(f'{OUTDIR}/dampingFrequency.png',dpi=300)


## Dump to ParaView
# Spatial modes
d.add_field('MODES_REAL',6,pyLOM.DMD.extract_modes(Phi,1,len(d),real=True,modes=[1,4,6,2,5,3]))
d.add_field('MODES_IMAG',6,pyLOM.DMD.extract_modes(Phi,1,len(d),real=False,modes=[1,4,6,2,5,3]))
pyLOM.io.pv_writer(m,d,'modes',basedir=f'{OUTDIR}/modes',instants=[0],times=[0.],vars=['MODES_REAL','MODES_IMAG'],fmt='vtkh5')

# Temporal evolution
d.add_field('RECON',len(VARIABLES),X_DMD)
pyLOM.io.pv_writer(m,d,'flow',basedir=f'{OUTDIR}/flow',instants=np.arange(t.shape[0],dtype=np.int32),times=t,vars=VARIABLES+['RECON'],fmt='vtkh5')


## Show and print timings
pyLOM.cr_info()
pyLOM.pprint(0,'End of output')


#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# Python interface for SPOD.
#
# Last rev: 02/09/2022
from __future__ import print_function

import numpy as np
import scipy

from ..utils.gpu import cp
from ..vmmath    import temporal_mean, subtract_mean, tsqr_svd, hammwin
from ..utils     import cr_nvtx as cr, cr_start, cr_stop


def _fft(Xf, winWeight, nDFT, nf):
	return (winWeight/nDFT)*scipy.fft.fft(Xf)[:nf]


## SPOD run method
@cr('SPOD.run')
def run(X, t, nDFT=0, nolap=0, remove_mean=True):
	'''
	Run SPOD analysis of a matrix X.

	Inputs:
		- X[ndims*nmesh,nt]: data matrix
		- dt:                timestep between adjacent snapshots
		- npwin:             number of points in each window (0 will set default value: ~10% nt)
		- nolap:             number of overlap points between windows (0 will set default value: 50% nwin)
		- remove_mean:       whether or not to remove the mean flow

	Returns:
		- L:  modal energy spectra.
		- P:  SPOD modes, whose spatial dimensions are identical to those of X.
		- f:  frequency vector.
	''' 
	cnp = cp if type(X) is cp.ndarray else np
	M,N = X.shape
	dt  = t[1] - t[0]
	cdtype = np.complex128 if X.dtype is np.double else np.complex64
	
	if nDFT == 0:
		nDFT = int(np.power(2,np.floor(np.log2(N/10))))
	window = hammwin(nDFT)
	if nolap == 0:
		nolap = int(np.floor(nDFT/2))
	nBlks = int(np.floor((N-nolap)/(nDFT-nolap)))
	# Correction for FFT window gain
	winWeight = 1/np.mean(window)

	# Remove temporal mean
	if remove_mean:
		cr_start('SPOD.temporal_mean',0)
		X_mean = temporal_mean(X)
		Y      = subtract_mean(X, X_mean)
		cr_stop('SPOD.temporal_mean',0)
	else:
		Y = X.copy()

	# Set frequency axis
	f  = np.arange(np.ceil(nDFT / 2) + 1) / dt / nDFT
	nf = f.shape[0]
	qk = np.zeros((M,nf),cdtype)
	Q  = np.zeros((M*nf,nBlks),cdtype)
	# Sent to CPU for FFT
	Y  = cp.asnumpy(Y) if type(X) is cp.ndarray else Y
	cr_start('SPOD.fft',0)
	for iblk in range(nBlks):
		# Get time index for present block
		i0 = iblk*(nDFT - nolap)
		ix = np.arange(nDFT) + i0
		for ip in range(M):
			Xf = Y[ip, ix].copy()*window
			qk[ip, :] = _fft(Xf, winWeight, nDFT, nf)
		qk[:,1:-1] *= 2
		Q[:, iblk] = qk.reshape((M*nf), order='F')
	cr_stop('SPOD.fft',0)

	Q  = cp.asarray(Q) if type(X) is cp.ndarray else Q
	L  = cnp.zeros((nf,nBlks),X.dtype)
	P  = cnp.zeros((M*nBlks,nf),X.dtype)
	cr_start('SPOD.SVD',0)
	for ifreq, freq in enumerate(f):
		qf         = Q[ifreq*M:(ifreq+1)*M, :].copy()/cnp.sqrt(nBlks)
		U, S, V    = tsqr_svd(qf)
		P[:,ifreq] = cnp.real(U.reshape((M*nBlks), order='F'))
		L[ifreq,:] = cnp.abs(S*S)
	cr_stop('SPOD.SVD',0)

	cr_start('SPOD.sort',0)
	f     = cp.asarray(f) if type(X) is cp.ndarray else f
	order = cnp.argsort(L[:,0])[::-1]
	P = P[:, order]
	f = f[order]
	L = L[order,:]
	cr_stop('SPOD.sort',0)
	  
	return L, P, f



#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# NN pipeline routines.
#
# Last rev: 02/10/2024

from typing         import List, Dict, Any
from .optimizer     import OptunaOptimizer
from ..utils.errors import raiseWarning
from ..            import pprint

class Pipeline:
    r"""
    Pipeline class to train and evaluate models. 
    To optimize a model, provide an optimizer and model class.
    To train a model with fixed parameters, provide a model and training parameters.

    Args: 
        train_dataset: The training dataset.
        valid_dataset (optional): The validation dataset. Default is ``None``.
        test_dataset (optional): The test dataset. Default is ``None``.
        model (Model, optional): The model to train. Default is ``None``. 
            If optimizer and model_class are provided, this is not used.
        training_params (Dict, optional): The parameters for training the model. Default is ``None``. 
            If optimizer and model_class are provided, this is not used.
        optimizer (OptunaOptimizer, optional): The optimizer to use for optimization. Default is ``None``.
        model_class (Model, optional): The model class to use for optimization. Default is ``None``.
        evaluators (List, optional): The evaluators to use for evaluating the model. Default is ``[]``.

    Raises:
        AssertionError: If neither model and training_params nor optimizer and model_class are provided.
    """

    def __init__(
        self,
        train_dataset,
        valid_dataset=None,
        test_dataset=None, 
        model=None, 
        training_params: Dict = None,
        optimizer: OptunaOptimizer = None,
        model_class=None,
        evaluators: List = [],
    ):
        self._model = model
        self.train_dataset = train_dataset
        self.test_dataset = test_dataset
        self.valid_dataset = valid_dataset
        self.optimizer = optimizer
        self.training_params = training_params
        self.model_class = model_class
        self.evaluators = evaluators

        assert (self.optimizer is not None and self.model_class is not None) or (
            self._model is not None and self.training_params is not None
        ), "Either model and training_params or optimizer and model_class must be provided"

    @property
    def model(self):
        """
        Get the trained model.
        """
        return self._model

    def run(self) -> Any:
        """
        Run the pipeline, this will train the model and return the output of the model's fit method. If optuna is used, the model will be trained with the optimized parameters.

        Returns:
            model_output (Any): The output of the model's fit method.
        """
        if self.optimizer is not None:
            if self.valid_dataset is None:
                self.valid_dataset = self.train_dataset
                raiseWarning( "Validation dataset not provided, using train dataset for evaluation on optimization")

            self._model, self.training_params = self.model_class.create_optimized_model(
                train_dataset = self.train_dataset,
                eval_dataset = self.valid_dataset,
                optuna_optimizer = self.optimizer,
            )
            pprint(0, "Training now a model with optimized parameters")

        model_output = self._model.fit(
            self.train_dataset, eval_dataset=self.test_dataset, **self.training_params
        )

        return model_output


#!/usr/bin/env cpython
#
# pyLOM - Python Low Order Modeling.
#
# Math operations module - statistics.
#
# Last rev: 27/10/2021

cimport cython
cimport numpy as np

import numpy as np

from .cfuncs    cimport c_sRMSE, c_sRMSE_relative, c_dRMSE, c_dRMSE_relative, c_sMAE, c_dMAE, c_sr2, c_dr2
from ..utils.cr  import cr

ctypedef fused realM:
	float[:]
	double[:]
	float[:,:]
	double[:,:]


@cr('math.RMSE')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def RMSE(realM A, realM B, int relative=True):
	'''
	Compute RMSE between A and B
	'''
	cdef int m = A.shape[0], n = A.shape[1] if A.ndim > 1 else 1
	if realM is double[:]:
		return c_dRMSE_relative(&A[0],&B[0],m,n) if relative else c_dRMSE(&A[0],&B[0],m,n)
	elif realM is double[:,:]:
		return c_dRMSE_relative(&A[0,0],&B[0,0],m,n) if relative else c_dRMSE(&A[0,0],&B[0,0],m,n)
	elif realM is float[:]:
		return c_sRMSE_relative(&A[0],&B[0],m,n) if relative else c_sRMSE(&A[0],&B[0],m,n)
	else:
		return c_sRMSE_relative(&A[0,0],&B[0,0],m,n) if relative else c_sRMSE(&A[0,0],&B[0,0],m,n)

@cr('math.MAE')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def MAE(realM A, realM B):
	'''
	Compute MAE between A and B
	'''
	cdef int m = A.shape[0], n = A.shape[1] if A.ndim > 1 else 1
	if realM is double[:]:
		return c_dMAE(&A[0],&B[0],m,n)
	elif realM is double[:,:]:
		return c_dMAE(&A[0,0],&B[0,0],m,n) 
	elif realM is float[:]:
		return c_sMAE(&A[0],&B[0],m,n)
	else:
		return c_sMAE(&A[0,0],&B[0,0],m,n)

@cr('math.r2')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def r2(realM A, realM B):
	'''
	Compute MAE between A and B
	'''
	cdef int m = A.shape[0], n = A.shape[1] if A.ndim > 1 else 1
	if realM is double[:]:
		return c_dr2(&A[0],&B[0],m,n)
	elif realM is double[:,:]:
		return c_dr2(&A[0,0],&B[0,0],m,n) 
	elif realM is float[:]:
		return c_sr2(&A[0],&B[0],m,n)
	else:
		return c_sr2(&A[0,0],&B[0,0],m,n)


import os
from typing import Dict, Tuple, Optional

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader

from ... import cr, pprint  # pyLOM/__init__.py
from .. import DEVICE  # pyLOM/NN/__init__.py
from ...utils.errors import raiseError, raiseWarning
from ..optimizer import OptunaOptimizer


class KAN(nn.Module):
    r"""
    KAN (Kolmogorov-Arnold Network) model for regression tasks. This model is based on https://arxiv.org/abs/2404.19756, inspired by the Kolmogorov-Arnold representation theorem.

    Args:
        input_size (int): The number of input features.
        output_size (int): The number of output features.
        n_layers (int): The number of hidden layers.
        hidden_size (int): The number of neurons in the hidden layers.
        layer_type (nn.Module): The type of layer to use in the model. It can be one of the following: ``JacobiLayer``, ``ChebyshevLayer``.
        model_name (str): The name of the model.
        p_dropouts (float, Optional): The dropout probability (default: ``0.0``).
        device (torch.device, Optional): The device where the model is loaded (default: gpu if available).
        verbose (bool, Optional): Whether to print the model information (default: ``True``).
        **layer_kwargs: Additional keyword arguments to pass to the layer type. For example, the order of the Taylor series or the degree of the Chebyshev polynomial.
    """

    def __init__(
        self,
        input_size: int,
        output_size: int,
        n_layers: int,
        hidden_size: int,
        layer_type,
        model_name: str = "KAN",
        p_dropouts: float = 0.0,
        device: torch.device = DEVICE,
        verbose: bool = True,
        **layer_kwargs,
    ):
        super().__init__()
        self.input_size = input_size
        self.output_size = output_size
        self.n_layers = n_layers
        self.hidden_size = hidden_size
        self.layer_type = layer_type
        self.model_name = model_name
        self.p_dropouts = p_dropouts
        self.device = device

        # Hidden layers with dropout
        hidden_layers = []
        for _ in range(n_layers):
            hidden_layers.append(layer_type(hidden_size, hidden_size, **layer_kwargs))
            hidden_layers.append(nn.Dropout(p=p_dropouts))

        self.kan_layers = nn.ModuleList(hidden_layers)

        # Input and output layers
        self.input = layer_type(input_size, hidden_size, **layer_kwargs)
        self.output = layer_type(hidden_size, output_size, **layer_kwargs)

        self.to(self.device)
        if verbose:
            pprint(0, f"Creating model KAN: {self.model_name}")
            keys_print = [
                "input_size",
                "output_size",
                "n_layers",
                "hidden_size",
                "layer_type",
                "p_dropouts",
                "device",
            ]
            for key in keys_print:
                pprint(0, f"\t{key}: {getattr(self, key)}")
            pprint(0,
                f"\ttotal_size (trained params):\t{sum(p.numel() for p in self.parameters() if p.requires_grad)}"
            )

    def forward(self, x):
        x = self.input(x)
        for layer in self.kan_layers:
            x = layer(x)

        x = self.output(x)

        return x

    @cr("KAN.fit")
    def fit(
        self,
        train_dataset: torch.utils.data.Dataset,
        eval_dataset: torch.utils.data.Dataset,
        batch_size: int = 32,
        epochs: int = 100,
        lr: float = 0.001,
        optimizer_class=optim.Adam,
        scheduler_type="StepLR",
        opti_kwargs={},
        lr_kwargs={},
        print_eval_rate: int = 2,
        loss_fn=nn.MSELoss(),
        save_logs_path=None,
        verbose: bool = True,
        max_norm_grad=float("inf"),
        **kwargs,
    ):
        r"""
        Train the model using the provided training dataset. The model is trained using the Adam optimizer with the provided learning rate and learning rate decay factor.

        Args:
            train_dataset (torch.utils.data.Dataset): The training dataset.
            eval_dataset (torch.utils.data.Dataset): The evaluation dataset.
            batch_size (int): The batch size. (default: ``32``).
            epochs (int): The number of epochs to train the model. (default: ``100``).
            lr (float): The learning rate for the Adam optimizer. (default: ``0.001``).
            optimizer_class (torch.optim, Optional): The optimizer to use. Available all optimizers from PyTorch except AdaDelta. (default: ``optim.Adam``).
            scheduler_type (str, opcional): Scheduler type to adjust the learning rate dynamically. (default: ``"StepLR"``).
                Available options:

                - "StepLR": Reduce the learning rate by a factor every ``step_size`` batches.
                - "ReduceLROnPlateau": Reduces the learning rate when a metric has stopped improving.
                - "OneCycleLR": Adjust the learning rate in a single cycle of the training.
            lr_kwargs (dict, opcional): Dictionary containing the specific parameters for the learning rate scheduler. (default: ``{}``).
                Some examples are:
                
                - StepLR: {"step_size": int, "gamma": float}.
                - ReduceLROnPlateau: {"mode": str, "factor": float, "patience": int}.
                - OneCycleLR: {"anneal_strategy": str, "div_factor": float}.
            opti_kwargs (dict, Optional): Additional keyword arguments to pass to the optimizer (default: `{}`).
            print_eval_rate (int, Optional): The model will be evaluated every ``print_eval_rate`` epochs and the losses will be printed. If set to 0, nothing will be printed (default: ``2``).
            loss_fn (torch.nn.Module, Optional): The loss function (default: ``nn.MSELoss()``).
            save_logs_path (str, Optional): Path to save the training and evaluation losses (default: ``None``).
            verbose (bool, Optional): Whether to print the training information (default: ``True``).
            max_norm_grad (float, Optional): The maximum norm of the gradients (default: ``float('inf')``).
            kwargs (dict, Optional): Additional keyword arguments to pass to the DataLoader. Can be used to set the parameters of the DataLoader (see PyTorch documentation at https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader):
               
                - batch_size (int, Optional): Batch size (default: ``32``).
                - shuffle (bool, Optional): Shuffle the data (default: ``True``).
                - num_workers (int, Optional): Number of workers to use (default: ``0``).
                - pin_memory (bool, Optional): Pin memory (default: ``True``).
        """
        if verbose:
            pprint(0, "")
            pprint(0, f"TRAINNING MODEL {self.model_name}")
            pprint(0, "")
            pprint(0, "Conditions:")
            pprint(0, f"\tepochs:     {epochs}")
            pprint(0, f"\tbatch size: 2**{int(np.log2(batch_size))}")
            pprint(0, f"\toptimizer class:  {optimizer_class}")
            pprint(0, f"\tscheduler:  {scheduler_type}")
            pprint(0, f"\tloss_fn:  {loss_fn}")
            pprint(0, f"\tsave_path:  {save_logs_path}")
            pprint(0, "\t")
            pprint(0, "Scheduler conditions:")
            for key, value in sorted(lr_kwargs.items()):
                if isinstance(value, dict):
                    pprint(0, f"\t{key}:")
                    for subkey, subvalue in sorted(value.items()):
                        pprint(0, f"\t{subkey}: {subvalue}")
                else:
                    pprint(0, f"\t{key}: {value}")
            pprint(0, "   ")
        dataloader_params = {
            "batch_size": batch_size,
            "shuffle": True,
            "num_workers": 0,
            "pin_memory": True,
        }
        for key in dataloader_params.keys():
            if key in kwargs:
                dataloader_params[key] = kwargs[key]
        train_loader = DataLoader(train_dataset, **dataloader_params)
        test_loader = DataLoader(eval_dataset, **dataloader_params)

        train_losses = torch.tensor([], device=self.device)
        test_losses = torch.tensor([], device=self.device)

        loss_iterations_train = []
        loss_iterations_test = []
        self.optimizer = optimizer_class(self.parameters(), lr=lr, **opti_kwargs)
        current_lr_vec = []

        if scheduler_type == "StepLR":
            self.scheduler = torch.optim.lr_scheduler.StepLR(self.optimizer, **lr_kwargs)
        elif scheduler_type == "ReduceLROnPlateau":
            self.scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
                self.optimizer, **lr_kwargs
            )
        elif scheduler_type == "OneCycleLR":
            self.scheduler = torch.optim.lr_scheduler.OneCycleLR(
                self.optimizer,
                max_lr=lr,
                steps_per_epoch=len(train_loader),
                epochs=1,
                **lr_kwargs,
            )
        else:
            raiseError(f"Invalid scheduler_type: {scheduler_type}. Available options are: 'StepLR', 'ReduceLROnPlateau', 'OneCycleLR'")

        if hasattr(self, "optimizer_state_dict"):
            self.optimizer.load_state_dict(self.optimizer_state_dict)
            del self.optimizer_state_dict

        if hasattr(self, "scheduler_state_dict"):
            self.scheduler.load_state_dict(self.scheduler_state_dict)
            del self.scheduler_state_dict

        for epoch in range(epochs):
            self.train()
            train_loss = 0.0

            def closure():
                self.optimizer.zero_grad()
                outputs = self(inputs)
                loss = loss_fn(outputs, targets)
                loss.backward()
                torch.nn.utils.clip_grad_norm_(
                    self.parameters(), max_norm=max_norm_grad
                )
                loss_iterations_train.append(loss.item())
                return loss

            for inputs, targets in train_loader:
                inputs, targets = (
                    inputs.float().to(self.device),
                    targets.float().to(self.device),
                )
                train_loss += self.optimizer.step(closure).item()
                if scheduler_type != "ReduceLROnPlateau":
                    self.scheduler.step()
                    current_lr = self.optimizer.param_groups[0]["lr"]
                    current_lr_vec.append(current_lr)

            train_loss /= len(train_loader)
            train_losses = torch.cat(
                (
                    train_losses,
                    torch.tensor([train_loss], dtype=torch.float64, device=self.device),
                )
            )
            if scheduler_type == "ReduceLROnPlateau":
                self.scheduler.step(train_loss)

            if (epoch + 1) % print_eval_rate == 0:
                self.eval()
                test_loss = 0.0
                with torch.no_grad():
                    for inputs, targets in test_loader:
                        inputs, targets = (
                            inputs.float().to(self.device),
                            targets.float().to(self.device),
                        )
                        outputs = self(inputs)
                        loss = loss_fn(outputs, targets)
                        loss_iterations_test.append(loss.item())
                        test_loss += loss.item()

                test_loss /= len(test_loader)
                test_losses = torch.cat(
                    (
                        test_losses,
                        torch.tensor(
                            [test_loss], dtype=torch.float64, device=self.device
                        ),
                    )
                )
                current_lr = self.optimizer.param_groups[0]["lr"]
                current_lr_vec.append(current_lr)
                if torch.cuda.is_available():
                    mem_used = torch.cuda.memory_allocated() / (1024**2)  # Memory usage in MB
                    memory_usage_str = f", MEM: {mem_used:.2f} MB"
                else:
                    memory_usage_str = ""

                if verbose:
                    pprint(
                        0,
                        f"Epoch {epoch + 1}/{epochs}, Train Loss: {train_loss:.4e}, Test Loss: {test_loss:.4e}, "
                        f"LR: {current_lr:.2e}{memory_usage_str}",
                    )

        if save_logs_path is not None:
            if not os.path.exists(save_logs_path):
                os.makedirs(save_logs_path)
            train_losses_np = train_losses.cpu().numpy()
            test_losses_np = test_losses.cpu().numpy()
            current_lr_np = np.array(current_lr_vec)
            if os.path.isdir(save_logs_path):
                if verbose:
                    pprint(0, f"Printing losses on path {save_logs_path}")
            else:
                if verbose:
                    pprint(0, "Path not found. Printing losses on local folder (.)")
                save_logs_path = "."

            np.save(
                save_logs_path + "/train_losses_" + self.model_name + ".npy",
                train_losses_np,
            )
            np.save(
                save_logs_path + "/test_losses_" + self.model_name + ".npy",
                test_losses_np,
            )
            np.save(
                save_logs_path + "/current_lr_" + self.model_name + ".npy",
                current_lr_np,
            )
            np.save(
                save_logs_path + "/losses_iterations_train_" + self.model_name + ".npy",
                np.array(loss_iterations_train),
            )
            np.save(
                save_logs_path + "/losses_iterations_test_" + self.model_name + ".npy",
                np.array(loss_iterations_test),
            )

        return {
            "train_loss": train_losses,
            "test_loss": test_losses,
            "lr": current_lr_vec,
            "loss_iterations_train": loss_iterations_train,
            "loss_iterations_test": loss_iterations_test,
        }

    @cr("KAN.predict")
    def predict(
        self,
        X: torch.utils.data.Dataset,
        return_targets: bool = False,
        **kwargs,
    ):
        r"""
        Predict the target values for the input data. The dataset is loaded into a DataLoader with the provided keyword arguments.
        The model is set to evaluation mode and the predictions are made using the input data. The output can be rescaled using
        the dataset scaler.

        Args:
            X (torch.utils.data.Dataset): The dataset whose target values are to be predicted using the input data.
            rescale_output (bool): Whether to rescale the output with the scaler of the dataset (default: ``True``).
            kwargs (dict, Optional): Additional keyword arguments to pass to the DataLoader. Can be used to set the parameters of the DataLoader (see PyTorch documentation at https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader):
                
                - **batch_size** (int, Optional): Batch size (default: ``32``).  
                - **shuffle** (bool, Optional): Shuffle the data (default: ``True``).  
                - **num_workers** (int, Optional): Number of workers to use (default: ``0``).  
                - **pin_memory** (bool, Optional): Pin memory (default: ``True``).  

        Returns:
            Tuple[np.ndarray, np.ndarray]: The predictions and the true target values.
        """

        dataloader_params = {
            "batch_size": 32,
            "shuffle": False,
            "num_workers": 0,
            "pin_memory": True,
        }

        for key in dataloader_params.keys():
            if key in kwargs:
                dataloader_params[key] = kwargs[key]

        predict_dataloader = DataLoader(X, **dataloader_params)

        total_rows = len(predict_dataloader.dataset)
        num_columns = self.output_size
        all_predictions = np.empty((total_rows, num_columns))
        all_targets = np.empty((total_rows, num_columns))

        self.eval()
        start_idx = 0
        with torch.no_grad():
            for x, y in predict_dataloader:
                output = self(x.to(self.device))
                batch_size = x.size(0)
                end_idx = start_idx + batch_size
                all_predictions[start_idx:end_idx, :] = output.cpu().numpy()
                all_targets[start_idx:end_idx, :] = y.cpu().numpy()
                start_idx = end_idx

        if return_targets:
            return all_predictions, all_targets
        else:
            return all_predictions

    def save(self, path: str, save_only_model: bool = False):
        r"""
        Save the model to a checkpoint file.

        Args:
            path (str): Path to save the model. It can be either a path to a directory or a file name.
            If it is a directory, the model will be saved with a filename that includes the number of epochs trained.
            save_only_model (bool, Optional): Whether to only save the model, or also the optimizer and scheduler. Note that when this is true, you won't be able to resume training from checkpoint.(default: ``False``).
        """
        checkpoint = {
            "input_size": self.input_size,
            "output_size": self.output_size,
            "n_layers": self.n_layers,
            "hidden_size": self.hidden_size,
            "layer_type": self.layer_type,
            "model_name": self.model_name,
            "p_dropouts": self.p_dropouts,
            "state_dict": self.state_dict(),
        }
        if hasattr(self.input, "degree"):
            checkpoint["degree"] = self.input.degree

        if not save_only_model:
            checkpoint["optimizer"] = self.optimizer.state_dict()
            checkpoint["scheduler"] = self.scheduler.state_dict()

        if os.path.isdir(path):
            filename = f"{self.model_name}.pth"
            path = path + filename

        torch.save(checkpoint, path)

    @classmethod
    def load(cls, path: str, device: torch.device = torch.device("cpu")):
        """
        Loads a model from a checkpoint file.

        Args:
            path (str): Path to the checkpoint file.
            device (torch.device): Device where the model is loaded (default: cpu).

        Returns:
            model (KAN): The loaded KAN model with the trained weights.
        """

        pprint(0, "Loading model...")
        checkpoint = torch.load(path, map_location=device, weights_only=False)
        raiseWarning("The model has been loaded with weights_only set to False. According with torch documentation, this is not recommended if you do not trust the source of your saved model, as it could lead to arbitrary code execution.")

        degree = checkpoint["degree"]
        layer_kwargs = {"degree": degree}

        model = cls(
            input_size=checkpoint["input_size"],
            output_size=checkpoint["output_size"],
            n_layers=checkpoint["n_layers"],
            hidden_size=checkpoint["hidden_size"],
            layer_type=checkpoint["layer_type"],
            model_name=checkpoint["model_name"],
            p_dropouts=checkpoint["p_dropouts"],
            device=device,
            **layer_kwargs,  # Pass the specific layer arguments
        )

        if "optimizer" in checkpoint:
            model.optimizer_state_dict = checkpoint["optimizer"]
        if "scheduler" in checkpoint:
            model.scheduler_state_dict = checkpoint["scheduler"]

        model.load_state_dict(checkpoint["state_dict"])
        pprint(0, f"Loaded KAN model: {checkpoint['model_name']}")
        keys_print = [
            "input_size",
            "output_size",
            "n_layers",
            "hidden_size",
            "layer_type",
            "p_dropouts",
        ]
        for key in keys_print:
            pprint(0, f"{key}: {checkpoint[key]}")

        return model

    @classmethod
    @cr("KAN.create_optimized_model")
    def create_optimized_model(
        cls,
        train_dataset: torch.utils.data.Dataset,
        eval_dataset: torch.utils.data.Dataset,
        optuna_optimizer: OptunaOptimizer,
        **kwargs,
    ) -> Tuple[nn.Module, Dict]:
        """
        Create an optimized KAN model using Optuna. The model is trained on the training dataset and the metric to optimize is computed with the evaluation dataset.
        If the parameters from the optimizer are a tuple, the function will optimize the parameter. If the parameter is a single value, it will be fixed during optimization.

        Args:
            train_dataset (torch.utils.data.Dataset): The training dataset.
            eval_dataset (torch.utils.data.Dataset): The evaluation dataset.
            optuna_optimizer (OptunaOptimizer): The optimizer to use for optimization.
            kwargs: Additional keyword arguments.

        Returns:
            Tuple [KAN, Dict]: The optimized model and the optimization parameters.

        Example:
            >>> from pyLOM.NN import KAN, OptunaOptimizer
            >>> # Split the dataset
            >>> train_dataset, eval_dataset = dataset.get_splits([0.8, 0.2])
            >>>
            >>> # Define the optimization parameters
            >>> optimization_params = {
            >>>     "lr": (0.00001, 0.1),
            >>>     "batch_size": (10, 64),
            >>>     "hidden_size": (10, 40), # optimizable parameter
            >>>     "n_layers": (1, 4),
            >>>     "print_eval_rate": 2,
            >>>     "epochs": 10, # non-optimizable parameter
            >>>    "lr_kwargs":{
            >>>        "gamma": (0.95, 0.99),
            >>>        "step_size": 7000
            >>>    },
            >>>     "model_name": "kan_test_optuna",
            >>>     'device': device,
            >>>     "layer_type": (pyLOM.NN.ChebyshevLayer, pyLOM.NN.JacobiLayer),
            >>>     "layer_kwargs": {
            >>>         "degree": (3, 10),
            >>>     },
            >>> }
            >>>
            >>> # Define the optimizer
            >>> optimizer = OptunaOptimizer(
            >>>     optimization_params=optimization_params,
            >>>     n_trials=5,
            >>>     direction="minimize",
            >>>     pruner=optuna.pruners.MedianPruner(n_startup_trials=5, n_warmup_steps=5, interval_steps=1),
            >>>     save_dir=None,
            >>> )
            >>>
            >>> # Create the optimized model
            >>> model, optimization_params = KAN.create_optimized_model(train_dataset, eval_dataset, optimizer)
            >>>
            >>> # Fit the model
            >>> model.fit(train_dataset, eval_dataset, **optimization_params)
        """
        optimizing_parameters = optuna_optimizer.optimization_params
        input_size, output_size = (
            train_dataset[0][0].shape[0],
            train_dataset[0][1].shape[0],
        )

        def optimization_function(trial) -> float:
            model_parameters, training_parameters = cls._sample_kan_parameters(
                trial, optimizing_parameters
            )
            model = cls(
                input_size=input_size,
                output_size=output_size,
                **model_parameters,
            )
            model_logs = model.fit(train_dataset, eval_dataset, **training_parameters)
            return model_logs["test_loss"][-1].item()

        best_params = optuna_optimizer.optimize(optimization_function)
        # Update the optimizing parameters with the best parameters found
        optimizing_parameters.update(best_params)
        if "layer_kwargs" in optimizing_parameters:
            del optimizing_parameters["layer_kwargs"]
        # Update the learning rate kwargs with the best parameters found
        if "lr_kwargs" in optimizing_parameters:
            for key in optimizing_parameters["lr_kwargs"]:
                if key in best_params:
                    optimizing_parameters["lr_kwargs"][key] = best_params[key]

        # Separate the model and training parameters from the best parameters found.
        # Note: now optimizing_parameters contains the best parameters found and does not contain any tuple,
        # so this is an easy wan to separate the training and model parameters
        model_parameters, training_parameters = cls._sample_kan_parameters(
            None, optimizing_parameters
        )
        model = cls(input_size=input_size, output_size=output_size, **model_parameters)
        return model, training_parameters

    @classmethod
    def _sample_kan_parameters(cls, trial, optimizing_parameters):
        training_parameters = {}
        model_parameters = {}
        mandatory_params = ["n_layers", "hidden_size", "layer_type"]
        for param in mandatory_params:
            if param in optimizing_parameters:
                model_parameters[param] = cls._suggest_value(trial, param, optimizing_parameters)
            else:
                raiseError(f"A value or range to optimize for the {param} must be provided")
        if "p_dropouts" in optimizing_parameters:
            model_parameters["p_dropouts"] = cls._suggest_float_value(
            trial, "p_dropouts", optimizing_parameters
            )
        else:
            model_parameters["p_dropouts"] = 0.0
        
        if "layer_kwargs" in optimizing_parameters:
            for key in optimizing_parameters["layer_kwargs"]:
                if key not in ["degree", "a", "b"]:
                    raiseError(f"Invalid key {key} in layer_kwargs")
                else:
                    model_parameters[key] = cls._suggest_value(
                        trial, key, optimizing_parameters["layer_kwargs"])
        elif "degree" in optimizing_parameters:
            model_parameters["degree"] = cls._suggest_int_value(
                trial, "degree", optimizing_parameters
            )
        else:
            raiseError("Layer kwargs with at least a key for degree must be provided")

        if "epochs" in optimizing_parameters:
            training_parameters["epochs"] = cls._suggest_int_value(
                trial, "epochs", optimizing_parameters
            )
        else:
            training_parameters["epochs"] = 100
        if "batch_size" in optimizing_parameters:
            training_parameters["batch_size"] = cls._suggest_int_value(
                trial, "batch_size", optimizing_parameters
            )
        else:
            training_parameters["batch_size"] = 32
        if "lr" in optimizing_parameters:
            training_parameters["lr"] = cls._suggest_float_value(
                trial, "lr", optimizing_parameters, log_scale=True
            )
        else:
            training_parameters["lr"] = 0.001

        if "max_norm_grad" in optimizing_parameters:
            training_parameters["max_norm_grad"] = cls._suggest_float_value(
                trial, "max_norm_grad", optimizing_parameters
            )
        else:
            training_parameters["max_norm_grad"] = float("inf")
            
        if "optimizer_class" in optimizing_parameters:
            training_parameters["optimizer_class"] = cls._suggest_categorical_value(
                trial, "optimizer_class", optimizing_parameters
            )
        else:
            training_parameters["optimizer_class"] = optim.Adam

        if "lr_kwargs" in optimizing_parameters:
            training_parameters["lr_kwargs"] = {}
            for key in optimizing_parameters["lr_kwargs"]:
                training_parameters["lr_kwargs"][key] = cls._suggest_value(
                    trial, key, optimizing_parameters["lr_kwargs"]
                )
        else:
            training_parameters["lr_kwargs"] = {}


        # non optimizing parameters
        for no_optimizing_param in ["save_logs_path", "print_eval_rate", "loss_fn", "opti_kwargs", "scheduler_type"]:
            if no_optimizing_param in optimizing_parameters:
                training_parameters[no_optimizing_param] = optimizing_parameters[
                    no_optimizing_param
                ]
                if isinstance(training_parameters[no_optimizing_param], tuple):
                    raiseError(f"Invalid value for {no_optimizing_param}. It is not an optimizable parameter.")

        for no_optimizing_param in ["device", "model_name"]:
            if no_optimizing_param in optimizing_parameters:
                model_parameters[no_optimizing_param] = optimizing_parameters[
                    no_optimizing_param
                ]

        return model_parameters, training_parameters
    
    def _suggest_value(
        trial, parameter_name, optimizing_parameters,
    ):
        if isinstance(optimizing_parameters[parameter_name], tuple):
            if isinstance(optimizing_parameters[parameter_name][0], int):
                return trial.suggest_int(
                    parameter_name, *optimizing_parameters[parameter_name]
                )
            elif isinstance(optimizing_parameters[parameter_name][0], float):
                return trial.suggest_float(
                    parameter_name, *optimizing_parameters[parameter_name]
                )
            else:
                return trial.suggest_categorical(
                    parameter_name, optimizing_parameters[parameter_name]
                )
        else:
            return optimizing_parameters[parameter_name]

    def _suggest_int_value(
        trial, parameter_name, optimizing_parameters, log_scale=False
    ):
        if isinstance(optimizing_parameters[parameter_name], tuple):
            return trial.suggest_int(
                parameter_name, *optimizing_parameters[parameter_name], log=log_scale
            )
        else:
            return optimizing_parameters[parameter_name]

    def _suggest_float_value(
        trial, parameter_name, optimizing_parameters, log_scale=False
    ):
        if isinstance(optimizing_parameters[parameter_name], tuple):
            return trial.suggest_float(
                parameter_name, *optimizing_parameters[parameter_name], log=log_scale
            )
        else:
            return optimizing_parameters[parameter_name]

    def _suggest_categorical_value(trial, parameter_name, optimizing_parameters):
        if isinstance(optimizing_parameters[parameter_name], tuple):
            return trial.suggest_categorical(
                parameter_name, optimizing_parameters[parameter_name]
            )
        else:
            return optimizing_parameters[parameter_name]


class ChebyshevLayer(nn.Module):
    """
    Chebyshev layer for KAN model.

    Args:
        input_dim (int): The number of input features.
        output_dim (int): The number of output features.
        degree (int): The degree of the Chebyshev polynomial.
    """

    def __init__(self, input_dim, output_dim, degree, **kwargs):
        super().__init__()
        self.inputdim = input_dim
        self.outdim = output_dim
        self.degree = degree

        self.cheby_coeffs = nn.Parameter(torch.empty(input_dim, output_dim, degree + 1))
        nn.init.normal_(self.cheby_coeffs, mean=0.0, std=1 / (input_dim * (degree + 1)))

    def forward(self, x):
        x = torch.reshape(x, (-1, self.inputdim))  # shape = (batch_size, inputdim)
        # Initialize Chebyshev polynomial tensors
        cheby = torch.ones(x.shape[0], self.inputdim, self.degree + 1, device=x.device)
        if self.degree > 0:
            cheby[:, :, 1] = x
        for i in range(2, self.degree + 1):
            cheby[:, :, i] = (
                2 * x * cheby[:, :, i - 1].clone() - cheby[:, :, i - 2].clone()
            )
        # Compute the Chebyshev interpolation
        y = torch.einsum(
            "bid,iod->bo", cheby, self.cheby_coeffs
        )  # shape = (batch_size, outdim)
        y = y.view(-1, self.outdim)
        return y


class JacobiLayer(nn.Module):
    """
    Jacobi layer for KAN model.

    Args:
        input_dim (int): The number of input features.
        output_dim (int): The number of output features.
        degree (int): The degree of the Jacobi polynomial.
        a (float, Optional): The first parameter of the Jacobi polynomial (default: ``1.0``).
        b (float, Optional): The second parameter of the Jacobi polynomial (default: ``1.0``).
    """

    def __init__(self, input_dim, output_dim, degree, a=1.0, b=1.0):
        super().__init__()
        self.inputdim = input_dim
        self.outdim = output_dim
        self.a = a
        self.b = b
        self.degree = degree

        self.jacobi_coeffs = nn.Parameter(
            torch.empty(input_dim, output_dim, degree + 1)
        )

        nn.init.normal_(
            self.jacobi_coeffs, mean=0.0, std=1 / (input_dim * (degree + 1))
        )

    def forward(self, x):
        x = torch.reshape(x, (-1, self.inputdim))  # shape = (batch_size, inputdim)
        # Initialize Jacobian polynomial tensors
        jacobi = torch.ones(x.shape[0], self.inputdim, self.degree + 1, device=x.device)
        if (self.degree > 0):  ## degree = 0: jacobi[:, :, 0] = 1 (already initialized) ; degree = 1: jacobi[:, :, 1] = x ; d
            jacobi[:, :, 1] = ((self.a - self.b) + (self.a + self.b + 2) * x) / 2
        for i in range(2, self.degree + 1):
            theta_k = (
                (2 * i + self.a + self.b)
                * (2 * i + self.a + self.b - 1)
                / (2 * i * (i + self.a + self.b))
            )
            theta_k1 = (
                (2 * i + self.a + self.b - 1)
                * (self.a * self.a - self.b * self.b)
                / (2 * i * (i + self.a + self.b) * (2 * i + self.a + self.b - 2))
            )
            theta_k2 = (
                (i + self.a - 1)
                * (i + self.b - 1)
                * (2 * i + self.a + self.b)
                / (i * (i + self.a + self.b) * (2 * i + self.a + self.b - 2))
            )
            jacobi[:, :, i] = (
                (theta_k * x + theta_k1) * jacobi[:, :, i - 1].clone()
                - theta_k2 * jacobi[:, :, i - 2].clone()
            )  # 2 * x * jacobi[:, :, i - 1].clone() - jacobi[:, :, i - 2].clone()
        # Compute the Jacobian interpolation
        y = torch.einsum(
            "bid,iod->bo", jacobi, self.jacobi_coeffs
        )  # shape = (batch_size, outdim)
        y = y.view(-1, self.outdim)
        return y



#!/usr/bin/env cpython
#
# pyLOM - Python Low Order Modeling.
#
# Math operations module - truncation.
#
# Last rev: 27/10/2021
from __future__ import print_function, division

import numpy as np

from ..utils.gpu import cp
from .maths      import vector_norm, vector_sum
from ..utils     import cr_nvtx as cr, mpi_reduce


def compute_truncation_residual(S, r):
	'''
	Compute the truncation residual.
	r must be a float precision (r<1) where:
		- r > 0: target residual
		- r < 0: fraction of cumulative energy to retain
	'''
	N = 0
	if r > 0:
		normS = vector_norm(S,0)
		for ii in range(S.shape[0]):
			accumulative = vector_norm(S,ii)/normS
			if accumulative < r: break
			N += 1
	else:
		r = abs(r)
		normS = vector_sum(S,0)
		accumulative = 0
		for ii in range(S.shape[0]):
			accumulative += S[ii]/normS
			N += 1		
			if accumulative > r: break
	return N

@cr('math.energy')
def energy(original, rec):
	'''
	Compute reconstruction energy as in:
	Eivazi, H., Le Clainche, S., Hoyas, S., & Vinuesa, R. (2022). 
	Towards extraction of orthogonal and parsimonious non-linear modes from turbulent flows. 
	Expert Systems with Applications, 202, 117038.
	https://doi.org/10.1016
	'''
	p = cp if type(original) is cp.ndarray else np
	# Compute local sums
	local_num = p.sum((original - rec) ** 2)
	local_den = p.sum(original ** 2)

	# Use Allreduce to compute global sums and make them available on all ranks
	global_num = mpi_reduce(local_num,op='sum',all=True)
	global_den = mpi_reduce(local_den,op='sum',all=True)

	# Compute Ek (this will be identical on all ranks)
	return 1 - global_num / global_den


#!/usr/bin/env cpython
#
# pyLOM - Python Low Order Modeling.
#
# Math operations module - FFT.
#
# Last rev: 27/10/2021

cimport cython
cimport numpy as np

import numpy as np

from libc.string cimport memcpy
from .cfuncs     cimport real, USE_FFTW3, c_sfft, c_dfft, c_snfft, c_dnfft, c_dhammwin
from ..utils.cr   import cr

@cr('math.hammwin')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def hammwin(int N):
	r'''
	Hamming windowing

	Args:
		N (int): Number of steps.

	Returns:
		numpy.ndarray: Hamming windowing.
	'''
	cdef np.ndarray[np.double_t,ndim=1] out = np.zeros((N,), dtype=np.double)
	c_dhammwin(&out[0],N)
	return out

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _sfft(float[:] t, float[:] y, int equispaced):
	'''
	Compute the fft of a signal y that is sampled at a
	constant timestep. Return the frequency and PSD
	'''
	cdef int n = y.shape[0]
	cdef float ts = t[1] - t[0], k_left
	cdef np.ndarray[np.float32_t,ndim=1]     x
	cdef np.ndarray[np.complex64_t,ndim=1] yf
	cdef np.ndarray[np.float32_t,ndim=1] f   = np.zeros((n,), dtype=np.float32)
	cdef np.ndarray[np.float32_t,ndim=1] PSD = np.zeros((n,), dtype=np.float32)
	memcpy(&f[0],&y[0],n*sizeof(float))
	if equispaced:
		c_sfft(&PSD[0],&f[0],ts,n)
	else:
		if USE_FFTW3:
			c_snfft(&PSD[0],&f[0],&t[0],n)
		else:
			import nfft
			# Compute sampling frequency
			k_left = (t.shape[0]-1.)/2.
			f      = (np.arange(t.shape[0],dtype=np.float32)-k_left)/t[n-1]
			# Compute power spectra using fft
			x   = -0.5 + np.arange(t.shape[0],dtype=np.float32)/t.shape[0]
			yf  = nfft.nfft_adjoint(x,y,len(t))
			PSD = np.real(yf*np.conj(yf))/y.shape[0]
	return f, PSD

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _dfft(double[:] t, double[:] y, int equispaced):
	'''
	Compute the fft of a signal y that is sampled at a
	constant timestep. Return the frequency and PSD
	'''
	cdef int n = y.shape[0]
	cdef double ts = t[1] - t[0], k_left
	cdef np.ndarray[np.double_t,ndim=1]     x
	cdef np.ndarray[np.complex128_t,ndim=1] yf
	cdef np.ndarray[np.double_t,ndim=1] f   = np.zeros((n,), dtype=np.double)
	cdef np.ndarray[np.double_t,ndim=1] PSD = np.zeros((n,), dtype=np.double)
	memcpy(&f[0],&y[0],n*sizeof(double))
	if equispaced:
		c_dfft(&PSD[0],&f[0],ts,n)
	else:
		if USE_FFTW3:
			c_dnfft(&PSD[0],&f[0],&t[0],n)
		else:
			import nfft
			# Compute sampling frequency
			k_left = (t.shape[0]-1.)/2.
			f      = (np.arange(t.shape[0],dtype=np.double)-k_left)/t[n-1]
			# Compute power spectra using fft
			x   = -0.5 + np.arange(t.shape[0],dtype=np.double)/t.shape[0]
			yf  = nfft.nfft_adjoint(x,y,len(t))
			PSD = np.real(yf*np.conj(yf))/y.shape[0]
	return f, PSD

@cr('math.fft')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def fft(real[:] t, real[:] y, int equispaced=True):
	r'''
	Compute the PSD of a signal y. For non equispaced time samples
	the nfft package is required.

	Args:
		t (numpy.ndarray): time vector.
		y (numpy.ndarray): signal vector.
		equispaced (bool): whether the samples in the time vector are equispaced or not.

	Returns:
		numpy.ndarray: frequency.
		numpy.ndarray: power density spectra.
	'''
	if real is double:
		return _dfft(t,y,equispaced)
	else:
		return _sfft(t,y,equispaced)


#!/usr/bin/env cpython
#
# pyLOM - Python Low Order Modeling.
#
# Math operations module - FFT.
#
# Last rev: 27/10/2021
from __future__ import print_function, division

import numpy as np, scipy, nfft

from .maths  import conj
from ..utils import cr_nvtx as cr


@cr('math.hammwin')
def hammwin(N:int) -> np.ndarray:
	r'''
	Hamming windowing

	Args:
		N (int): Number of steps.

	Returns:
		numpy.ndarray: Hamming windowing.
	'''
	return np.transpose(0.54-0.46*np.cos(2*np.pi*np.arange(N)/(N-1)))

@cr('math.fft')
def fft(t:np.ndarray,y:np.ndarray,equispaced:bool=True) -> np.ndarray:
	r'''
	Compute the PSD of a signal y. For non equispaced time samples
	the nfft package is required.

	Args:
		t (numpy.ndarray): time vector.
		y (numpy.ndarray): signal vector.
		equispaced (bool): whether the samples in the time vector are equispaced or not.

	Returns:
		numpy.ndarray: frequency.
		numpy.ndarray: power density spectra.
	'''
	if equispaced:
		ts = t[1] - t[0] # Sampling time
		# Compute sampling frequency
		f  = 1./ts/t.shape[0]*np.arange(t.shape[0],dtype=y.dtype)
		# Compute power spectra using fft
		yf = scipy.fft.fft(y)
	else:
		# Compute sampling frequency
		k_left = (t.shape[0]-1.)/2.
		f      = (np.arange(t.shape[0],dtype=y.dtype)-k_left)/t[-1]
		# Compute power spectra using fft
		x  = -0.5 + np.arange(t.shape[0],dtype=y.dtype)/t.shape[0]
		yf = nfft.nfft_adjoint(x,y,len(t))
	ps = np.real(yf*conj(yf))/y.shape[0] # np.abs(yf)/y.shape[0]
	return f, ps


#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# NN utility routines.
#
# Last rev: 02/10/2024

import os, random, json, numpy as np, torch
import torch.nn.functional as F

from torch.utils.data import Subset
from torch            import Generator, randperm, default_generator
from itertools        import product, accumulate
from typing           import List, Optional, Tuple, cast, Sequence, Union, Callable
from tqdm             import tqdm

from .                import DEVICE
from ..utils.cr       import cr
from ..utils.errors   import raiseWarning, raiseError
from ..dataset        import Dataset as pyLOMDataset


class MinMaxScaler:
    r"""
    Min-max scaling to scale variables to a desired range. The formula is given by:

    .. math::

        X_{scaled} = \frac{X - X_{min}}{X_{max} - X_{min}} * (feature\_range_{max} - feature\_range_{min}) + feature\_range_{min}

    Args:
        feature_range (Tuple): Desired range of transformed data. Default is ``(0, 1)``.
    """

    def __init__(self, feature_range=(0, 1)):
        self.feature_range = feature_range
        self._is_fitted = False

    @property
    def is_fitted(self):
        return self._is_fitted
    
    def fit(self, variables: Union[List[Union[np.ndarray, torch.tensor]], np.ndarray, torch.tensor]):
        """
        Compute the min and max values of each variable.
        Args:
            variables (List): List of variables to be fitted. The variables should be numpy arrays or torch tensors.
            A numpy array or torch tensor can be passed directly and each column will be considered as a variable to be scaled.
        """
        variables = self._cast_variables(variables)
        min_max_values = []
        for variable in variables:
            min_val = float(variable.min())
            max_val = float(variable.max())
            min_max_values.append({"min": min_val, "max": max_val})
        self.variable_scaling_params = min_max_values
        self._is_fitted = True

    def transform(
        self, variables: Union[List[Union[np.ndarray, torch.tensor]], np.ndarray, torch.tensor]
    ) -> List[Union[np.ndarray, torch.tensor]]:
        """
        Scale variables to the range defined on `feature_range` using min-max scaling.
        Args:
            variables (List): List of variables to be scaled. The variables should be numpy arrays or torch tensors.
            A numpy array or torch tensor can be passed directly and each column will be considered as a variable to be scaled.
        Returns:
            scaled_variables: List of scaled variables.
        """
        is_array = isinstance(variables, (np.ndarray))
        is_tensor = isinstance(variables, (torch.Tensor))
        variables = self._cast_variables(variables)
        scaled_variables = []
        for i, variable in enumerate(variables):
            min_val = self.variable_scaling_params[i]["min"]
            max_val = self.variable_scaling_params[i]["max"]
            data_range = max_val - min_val
            data_range = 1 if data_range == 0 else data_range
            scaled_variable = (variable - min_val) / data_range
            # scale the variable to the desired feature_range
            scaled_variable = (
                scaled_variable * (self.feature_range[1] - self.feature_range[0])
                + self.feature_range[0]
            )
            scaled_variables.append(scaled_variable)

        if is_array:
            scaled_variables = np.hstack(scaled_variables)
        elif is_tensor:
            scaled_variables = torch.hstack(scaled_variables)

        return scaled_variables
    
    def fit_transform(self, variables: List[Union[np.ndarray, torch.tensor]]) -> List[Union[np.ndarray, torch.tensor]]:
        """
        Fit and transform the variables using min-max scaling.
        Args:
            variables (List): List of variables to be fitted and scaled. The variables should be numpy arrays or torch tensors.
            A numpy array or torch tensor can be passed directly and each column will be considered as a variable to be scaled.
        Returns:
            scaled_variables: List of scaled variables.
        """
        self.fit(variables)
        return self.transform(variables)

    def inverse_transform(self, variables: List[Union[np.ndarray, torch.tensor]]) -> List[Union[np.ndarray, torch.tensor]]:
        """
        Inverse scale variables that have been scaled using min-max scaling.
        Args:
            variables (List): List of variables to be inverse scaled. The variables should be numpy arrays or torch tensors.
            A numpy array or torch tensor can be passed directly and each column will be considered as a variable to be scaled.
        Returns:
            inverse_scaled_variables: List of inverse scaled variables.
        """
        is_array = isinstance(variables, (np.ndarray))
        is_tensor = isinstance(variables, (torch.Tensor))
        variables = self._cast_variables(variables)
        if len(variables) != len(self.variable_scaling_params):
            raiseError(
                f"Number of variables to inverse transform ({len(variables)}) does not match the number of variables fitted ({len(self.variable_scaling_params)})"
            )
        inverse_scaled_variables = []
        for variable, scaling_params in zip(variables, self.variable_scaling_params):
            min_val = scaling_params["min"]
            max_val = scaling_params["max"]
            inverse_scaled_variable = (variable - self.feature_range[0]) / (
                self.feature_range[1] - self.feature_range[0]
            )
            data_range = max_val - min_val
            data_range = 1 if data_range == 0 else data_range
            inverse_scaled_variable = inverse_scaled_variable * data_range + min_val
            inverse_scaled_variables.append(inverse_scaled_variable)
        if is_array:
            inverse_scaled_variables = np.hstack(inverse_scaled_variables)
        elif is_tensor:
            inverse_scaled_variables = torch.hstack(inverse_scaled_variables)

        return inverse_scaled_variables

    def save(self, filepath: str) -> None:
        """
        Save the fitted scaler parameters to a JSON file.
        
        Args:
            filepath (str): Path where the scaler parameters will be saved
        """
        if not self.is_fitted:
            raiseError("Scaler must be fitted before it can be saved")
        
        save_dict = {
            "feature_range": self.feature_range,
            "variable_scaling_params": self.variable_scaling_params,
        }
        
        with open(filepath, 'w') as f:
            json.dump(save_dict, f, indent=4)

    @staticmethod
    def load(filepath: str) -> 'MinMaxScaler':
        """
        Load a saved MinMaxScaler from a JSON file.
        
        Args:
            filepath (str): Path to the saved scaler parameters
            
        Returns:
            MinMaxScaler: A new MinMaxScaler instance with the loaded parameters

        """
        if not os.path.exists(filepath):
            raiseError(f"No file found at {filepath}")
        
        with open(filepath, 'r') as f:
            loaded_dict = json.load(f)
        
        scaler = MinMaxScaler(feature_range=tuple(loaded_dict["feature_range"]))
        
        # Restore the scaling parameters
        scaler.variable_scaling_params = loaded_dict["variable_scaling_params"]
        scaler._is_fitted = True
        
        return scaler   

    def _cast_variables(self, variables):
        if isinstance(variables, (torch.Tensor)):
            variables = [variables[:, i].unsqueeze(1) for i in range(variables.shape[1])]
        elif isinstance(variables, (np.ndarray)):
            variables = [np.expand_dims(variables[:, i], axis=1) for i in range(variables.shape[1])]
        return variables


class Dataset(torch.utils.data.Dataset):
    r"""
    Dataset class to be used with PyTorch. It can be used with both mesh and point data.
    It is useful convert the `pyLOM.Dataset` to a PyTorch dataset and train neural networks with results from CFD simulations.

    Example:
        >>> original_dataset = pyLOM.Dataset.load(path)
        >>> input_scaler = pyLOM.NN.MinMaxScaler()
        >>> output_scaler = pyLOM.NN.MinMaxScaler()
        >>> dataset = pyLOM.NN.Dataset(
        ...     variables_out=(original_dataset["CP"],),
        ...     variables_in=original_dataset.xyz,
        ...     parameters=[[*zip(original_dataset.get_variable('AoA'), original_dataset.get_variable('Mach'))]], # to have each Mach and AoA pair just once. To have all possibnle combinations, use [original_dataset.get_variable('AoA'), original_dataset.get_variable("Mach")]
        ...     inputs_scaler=input_scaler,
        ...     outputs_scaler=output_scaler,
        ... )


    Args:
        variables_out (Tuple): Tuple of variables to be used as output. Each variable should be a 2d numpy array or torch tensor. If only one variable wants to be provided, it should be passed as a tuple with one element. E.g. ``(variable,)``.
        mesh_shape (Tuple): Shape of the mesh. If not mesh is used and the data is considered as points, leave this as default. Default is ``(1,)``.
        variables_in (np.ndarray): Input variables. Default is ``None``.
        parameters (List[List[float]]): List of parameters to be used as input. All possible combinations of these parameters, i.e. its cartesian product, will appear along with variables_in. If there is only one inner list and its elements are tuples, they will be treated as a single element for the cartesiand produt, which is useful when the combination of the parameters was predefined. Default is ``None``.
        inputs_scaler (MinMaxScaler): Scaler to scale the input variables. If the scaler is not fitted, it will be fitted. Default is ``None``.
        outputs_scaler (MinMaxScaler): Scaler to scale the output variables. If the scaler is not fitted, it will be fitted. Default is ``None``.
        snapshots_by_column (bool): If the snapshots from `variables_out` are stored by column. The snapshots on `pyLOM.Dataset`s have this format. Default is ``True``.
    """
    @cr("NN.Dataset.__init__")
    def __init__(
        self,
        variables_out: Tuple,
        mesh_shape: Tuple = (1,),
        variables_in: np.ndarray = None,
        parameters: List[List[float]] = None,
        combine_parameters_with_cartesian_prod: bool = False,
        inputs_scaler=None,
        outputs_scaler=None,
        snapshots_by_column=True, 
    ):
        self.parameters = parameters
        self.num_channels = len(variables_out)
        self.mesh_shape = mesh_shape
        if snapshots_by_column:
            variables_out = [variable.T for variable in variables_out]
        if outputs_scaler is not None:
            if not outputs_scaler.is_fitted:
                outputs_scaler.fit(variables_out)
            variables_out = outputs_scaler.transform(variables_out)
        self.variables_out = self._process_variables_out(variables_out)
        if variables_in is not None:
            self.parameters = self._process_parameters(parameters, combine_parameters_with_cartesian_prod)
            self.variables_in = torch.tensor(variables_in, dtype=torch.float32)
            if inputs_scaler is not None:
                variables_in_columns = [self.variables_in[:, i] for i in range(self.variables_in.shape[1])]
                parameters_columns = [self.parameters[:, i] for i in range(self.parameters.shape[1])] if self.parameters is not None else []
                if not inputs_scaler.is_fitted:
                    inputs_scaler.fit(variables_in_columns + parameters_columns)
                input_data_transformed = inputs_scaler.transform(variables_in_columns + parameters_columns)
                self.variables_in = torch.stack(input_data_transformed[:self.variables_in.shape[1]], dim=1).float()
                if self.parameters is not None:
                    self.parameters = torch.stack(input_data_transformed[self.variables_in.shape[1]:], dim=1)
        else:
            self.variables_in = None
            if parameters is not None:
                raiseWarning("Parameters were passed but no input variables were passed. Parameters will be ignored.")
            self.parameters = None

    def _process_variables_out(self, variables_out):
        variables_out_stacked = []
        for variable in variables_out:
            variable = torch.tensor(variable)
            variable = variable.reshape(-1, *self.mesh_shape)
            variables_out_stacked.append(variable)
        variables_out_stacked = torch.stack(variables_out_stacked, dim=1)

        if variables_out_stacked.shape[-1] == 1:  # (N, C, 1) -> (N, C)
            variables_out_stacked = variables_out_stacked.squeeze(-1)
        return variables_out_stacked.float()

    def _process_parameters(self, parameters, combine_parameters_with_cartesian_prod):
        if parameters is None:
            return None
        if not combine_parameters_with_cartesian_prod:
            # assert that all elements in the inner lists have the same length
            if not all(len(parameters[0]) == len(inner_list) for inner_list in parameters):
                raiseError("All parameter lists must have the same length")
            parameters = torch.tensor(list(zip(*parameters)))
            
        else:
            if len(parameters) == 1:
                parameters = torch.tensor(parameters[0])
            else:
                parameters = torch.tensor(list(product(*parameters)))
        return parameters.float()

    @property
    def shape(self):
        return self.variables_out.shape
    
    @property
    def num_parameters(self):
        return self.parameters.shape[1] if self.parameters is not None else 0

    def __len__(self):
        return len(self.variables_out)

    def __getitem__(self, idx: Union[int, slice]):
        """
        Return the input data and the output data for a given index as a tuple. If there is no input data, only the output data will be returned.
        If parameters are used, the parameters will be concatenated with the input data at the end.

        Args:
            idx (int, slice): Index of the data to be returned.
        """            
        if self.variables_in is None:
            return self.variables_out[idx]
        
        if isinstance(idx, slice):
            idx = torch.arange(
                idx.start if idx.start is not None else 0, 
                idx.stop if idx.stop is not None else len(self),
                idx.step if idx.step is not None else 1
            )
        else:
            idx = torch.tensor(idx) if not isinstance(idx, torch.Tensor) else idx
        
        variables_in_idx = idx % len(self.variables_in)
        parameters_idx = idx // len(self.variables_in)
        input_data = torch.hstack([self.variables_in[variables_in_idx], self.parameters[parameters_idx]]) if self.parameters is not None else self.variables_in[variables_in_idx]
        return input_data, self.variables_out[idx]


    def __setitem__(self, idx: Union[int, slice], value: Tuple):
        """
        Set the input data and the output data for a given index. If there is no input data, only the output data will be set.
        If parameters are used, the parameters will be concatenated with the input data at the end.

        Args:
            idx (int, slice): Index of the data to be set.
            value (Tuple): Tuple with the input data and the output data. If there is no input data, the tuple should have only one element.
        """
        if self.variables_in is None:
            self.variables_out[idx] = value
        else:
            if len(value) != 2:
                raiseError("Invalid value to set. Must be a tuple with two elements when variables_in takes some value")
            if isinstance(idx, slice):
                idx = torch.arange(
                    idx.start if idx.start is not None else 0, 
                    idx.stop if idx.stop is not None else len(self),
                    idx.step if idx.step is not None else 1
                )
            else:
                idx = torch.tensor(idx) if not isinstance(idx, torch.Tensor) else idx
            input_data, variables_out = value
            variables_in_idx = idx % len(self.variables_in)
            parameters_idx = idx // len(self.variables_in)

            self.variables_in[variables_in_idx] = input_data[:, :self.variables_in.shape[1]]
            if self.parameters is not None:
                self.parameters[parameters_idx] = input_data[:, self.variables_in.shape[1]:]
            self.variables_out[idx] = variables_out

        
    def __add__(self, other):
        """
        Concatenate two datasets. The datasets must have the same number of input coordinates and parameters.

        Args:
            other (Dataset): Dataset to be concatenated with.
        """
        if not isinstance(other, Dataset):
            raiseError(f"Cannot add Dataset with {type(other)}")
        if self.variables_in is not None:
            if other.variables_in is None or other.variables_in.shape[1] != self.variables_in.shape[1]:
                raiseError(f"Cannot add datasets with different number of columns on input coordinates, got {other.variables_in.shape[1]} and {self.variables_in.shape[1]}")
            if self.parameters is not None:
                if other.parameters is None or other.parameters.shape[1] != self.parameters.shape[1]:
                    raiseError("Cannot add datasets with different number of parameters")
                if other.variables_in is None or other.variables_in.shape[0] != self.variables_in.shape[0]:
                    raiseError(f"Cannot add datasets with different number of input coordinates, got {other.variables_in.shape[0]} and {self.variables_in.shape[0]}")
                self.parameters = torch.cat([self.parameters, other.parameters], dim=0)
            else:
                self.variables_in = torch.cat([self.variables_in, other.variables_in], dim=0)

        variables_out = torch.cat([self.variables_out, other.variables_out], dim=0)
        self.variables_out = variables_out
        return self
    
    def concatenate(self, other):
        """
        Alias for the `__add__` method.

        Args:
            other (Dataset): Dataset to be concatenated with.
        """
        return self.__add__(other)

    def _crop2D(self, nh, nw):
        self.variables_out = self.variables_out[:, :, :nh, :nw]
        self.mesh_shape    = (nh, nw)

    def _crop3D(self, nh, nw, nd):
        self.variables_out = self.variables_out[:, :, :nh, :nw, :nd]
        self.mesh_shape    = (nh, nw, nd)

    def crop(self, *args):
        """
        Crop the dataset to a desired shape. The cropping currently works for 2D and 3D meshes.

        Args:
            args (Tuple): Desired shape of the mesh. If the mesh is 2D, the shape should be a tuple with two elements. If the mesh is 3D, the shape should be a tuple with three elements.
        """
        if len(args) == 2: 
            self._crop2D(*args)
        elif len(args) == 3:
            self._crop3D(*args)
        else:
            raiseError(f'Invalid number of dimensions {len(args)} for mesh {self.mesh_shape}')

    def _pad2D(self, n0h, n0w):
        nh, nw = self.mesh_shape
        self.variables_out = F.pad(self.variables_out, (0, n0w-nw, 0, n0h-nh), mode='constant', value=0)
        self.mesh_shape    = (n0h,n0w)

    def _pad3D(self, n0h, n0w, n0d):
        nh, nw, nd = self.mesh_shape
        self.variables_out = F.pad(self.variables_out, (0, n0d-nd, 0, n0w-nw, 0, n0h-nh), mode='constant', value=0)
        self.mesh_shape    = (n0h, n0w, n0d)

    def pad(self, *args):
        """
        Pad the dataset to a desired shape. The padding currently works for 2D and 3D meshes.

        Args:
            args (Tuple): Desired shape of the mesh. If the mesh is 2D, the shape should be a tuple with two elements. If the mesh is 3D, the shape should be a tuple with three elements.
        """
        if len(args) == 2: 
            self._pad2D(*args)
        elif len(args) == 3:
            self._pad3D(*args)
        else:
            raiseError(f'Invalid number of dimensions {len(args)} for mesh {self.mesh_shape}')

    def get_splits(
        self,
        sizes: Sequence[int],
        shuffle: bool = True,
        return_views: bool = True,
        generator: Optional[Generator] = default_generator,
    ):
        """
        Randomly split the dataset into non-overlapping new datasets of given lengths.

        If a list of fractions that sum up to 1 is given,
        the lengths will be computed automatically as
        floor(frac * len(self)) for each fraction provided.

        After computing the lengths, if there are any remainders, 1 count will be
        distributed in round-robin fashion to the lengths
        until there are no remainders left.

        Optionally fix the generator for reproducible results.

        Args:
            sizes (sequence): lengths or fractions of splits to be produced.
            shuffle (bool): whether to shuffle the data before splitting. Default: ``True``.
            return_views (bool): If True, the splits will be returned as views of the original dataset in form of torch.utils.data.Subset.
            If False, the splits will be returned as new Dataset instances. Default: ``True``.
            Warning: If return_views is True, the original dataset must be kept alive, otherwise the views will point to invalid memory.
            Be careful when using `return_views=False` with `variables_in` and `parameters`, the memory usage will be higher.
            generator (Generator): Generator used for the random permutation. Default: ``default_generator``.
        
        Returns:
            List[Dataset | Subset]: List with the splits of the dataset.
        """
        if np.isclose(sum(sizes), 1) and sum(sizes) <= 1:
            subset_lengths: List[int] = []
            for i, frac in enumerate(sizes):
                if frac < 0 or frac > 1:
                    raise ValueError(f"Fraction at index {i} is not between 0 and 1")
                n_items_in_split = int(
                    np.floor(len(self) * frac)  # type: ignore[arg-type]
                )
                subset_lengths.append(n_items_in_split)
            remainder = len(self) - sum(subset_lengths)  # type: ignore[arg-type]
            # add 1 to all the lengths in round-robin fashion until the remainder is 0
            for i in range(remainder):
                idx_to_add_at = i % len(subset_lengths)
                subset_lengths[idx_to_add_at] += 1
            sizes = subset_lengths
            for i, length in enumerate(sizes):
                if length == 0:
                    raiseWarning(
                        f"Length of split at index {i} is 0. "
                        f"This might result in an empty dataset."
                    )

        # Cannot verify that dataset is Sized
        if sum(sizes) != len(self):  # type: ignore[arg-type]
            raiseError(
                "Sum of input lengths does not equal the length of the dataset!"
            )
        if shuffle:
            indices = randperm(sum(sizes), generator=generator).tolist()  # type: ignore[arg-type, call-overload]
        else:
            indices = list(range(sum(sizes)))
        sizes = cast(Sequence[int], sizes)

        split_datasets = []
        for offset, length in zip(accumulate(sizes), sizes):
            new_indices = indices[offset - length : offset]
            split_datasets.append(self._create_subset_from_indices(new_indices, return_views))
        return split_datasets

    def get_splits_by_parameters(
        self, 
        sizes: Sequence[int], 
        shuffle: bool = True, 
        return_views: bool = True,
        generator: Optional[Generator] = default_generator
    ):
        """
        Split the dataset into non-overlapping new datasets with diferent sets of parameters of given length.

        If a list of fractions that sum up to 1 is given,
        the lengths will be computed automatically as
        floor(frac * len(self.parameters)) for each fraction provided.

        After computing the lengths, if there are any remainders, 1 count will be
        distributed in round-robin fashion to the lengths
        until there are no remainders left.

        Optionally fix the generator for reproducible results.

        Args:
            sizes (sequence): lengths or fractions of splits to be produced
            shuffle (bool): whether to shuffle the data before splitting. Default: ``True``
            return_views (bool): If True, the splits will be returned as views of the original dataset in form of torch.utils.data.Subset.
            If False, the splits will be returned as new Dataset instances. Default: ``True``.
            Warning: If return_views is True, the original dataset must be kept alive, otherwise the views will point to invalid memory.
            generator (Generator): Generator used for the random permutation. Default: ``default_generator``.

        Returns:
            List[Dataset | Subset]: List with the splits of the dataset.
        """
        if self.parameters is None:
            raiseError("This Dataset does not have parameters to split by")
        if np.isclose(sum(sizes), 1) and sum(sizes) <= 1:
            subset_lengths: List[int] = []
            for i, frac in enumerate(sizes):
                if frac < 0 or frac > 1:
                    raise ValueError(f"Fraction at index {i} is not between 0 and 1")
                n_items_in_split = int(
                    np.floor(len(self.parameters) * frac)  # type: ignore[arg-type]
                )
                subset_lengths.append(n_items_in_split)
            remainder = len(self.parameters) - sum(subset_lengths)
            # add 1 to all the lengths in round-robin fashion until the remainder is 0
            for i in range(remainder):
                idx_to_add_at = i % len(subset_lengths)
                subset_lengths[idx_to_add_at] += 1
            sizes = subset_lengths
            for i, length in enumerate(sizes):
                if length == 0:
                    raiseWarning(
                        f"Length of split at index {i} is 0. "
                        f"This might result in an empty dataset."
                    )

        # Cannot verify that dataset is Sized
        if sum(sizes) != len(self.parameters):
            raiseError(
                "Sum of input lengths does not equal the length of the number of parameters!"
            )
        if shuffle:
            indices = randperm(sum(sizes), generator=generator).tolist()  # type: ignore[arg-type, call-overload]
        else:
            indices = list(range(sum(sizes)))
        sizes = cast(Sequence[int], sizes)
        split_datasets = []
        for offset, length in zip(accumulate(sizes), sizes):
            new_indices = []
            for index in indices[offset - length : offset]:
                vars_in_extended_indices = list(range(index * len(self.variables_in), (index + 1) * len(self.variables_in)))
                new_indices.extend(vars_in_extended_indices)
            if not return_views:
                # Get the split data
                if self.variables_in is not None:
                    split_variables_in = self.variables_in
                    split_parameters = self.parameters[indices[offset - length : offset]]
                else:
                    split_variables_in = None
                    split_parameters = None
                    
                # Get the split output variables
                split_variables_out = self.variables_out[new_indices]

                # Create a new dataset instance with the split data
                split_dataset = Dataset(
                    variables_out=tuple([split_variables_out[:, i].reshape(-1, *self.mesh_shape).numpy()
                                    for i in range(self.num_channels)]),
                    mesh_shape=self.mesh_shape,
                    variables_in=split_variables_in.numpy() if split_variables_in is not None else None,
                    # Parameters must be transposed to pass only as many list as the number of variables
                    parameters=split_parameters.T.tolist() if split_parameters is not None else None,
                    combine_parameters_with_cartesian_prod=False,
                    snapshots_by_column=False  # Since the data is already reshaped
                )
                split_datasets.append(split_dataset)
            else:
                split_datasets.append(Subset(self, new_indices))

        return split_datasets
    
    @classmethod
    def load(
        cls,
        file_path,
        field_names: List[str],
        variables_names: List[str] = ['all'],
        add_mesh_coordinates: bool = True,
        **kwargs,
    ):
        """
        Create a Dataset from a saved `pyLOM.Dataset` in one of its formats.

        Args:
            file_path (str): Path to the HDF5 file.
            variables_out_names (List[str]): Names of the fields to be used as output. E.g. ``["CP"]``.
            add_variables (bool): Whether to add the variables as input variables. Default is ``True``.
            variables_names (List[str]): Names of the variables from pyLOM.Dataset.varnames to be used as input. If ``["all"]`` is passed, all variables will be used. Default is ``["all"]``.
            kwargs: Additional arguments to be passed to the pyLOM.NN.Dataset constructor.

        Returns:
            Dataset: Dataset created from the saved `pyLOM.Dataset`.

        Example:
            >>> dataset = pyLOM.NN.Dataset.load(
            ...     file_path,
            ...     field_names=["CP"],
            ...     add_variables=True,
            ...     add_mesh_coordinates=True,
            ...     inputs_scaler=inputs_scaler,
            ...     outputs_scaler=outputs_scaler,
            ... )
        """
        
        original_dataset = pyLOMDataset.load(file_path)
        if (len(variables_names) >= 0 and "combine_parameters_with_cartesian_prod" in kwargs):
            raiseWarning(
                """Be careful when using combine_parameters_with_cartesian_prod and passing variables_names.
                You need to make sure that the length of cartesian product multiplied by the length 
                of the mesh coordinates is equal to the length of the fields on field_names."""
            )
        if variables_names == ["all"]:
            if not add_mesh_coordinates:
                raiseError("Cannot use all variables without adding mesh coordinates")
            if len(original_dataset.varnames) == 0:
                raiseError("No variabele found in the dataset")
            variables_names = original_dataset.varnames

        parameters = [original_dataset.get_variable(var_name) for var_name in variables_names]

        return cls(
            variables_out=tuple(
                [original_dataset[var_name] for var_name in field_names]
            ),
            parameters=parameters if len(parameters) > 0 else None,
            variables_in=original_dataset.xyz if add_mesh_coordinates else None,
            **kwargs,
        )
    
    @cr("NN.Dataset.map")
    def map(
        self,
        function: Callable,
        fn_kwargs: dict = {},
        batched: bool = False,
        batch_size: int = 1000,
    ):
        """
        Apply a function to the dataset.

        Args:
            function (Callable): Function to be applied to the dataset with one of the following signatures:
                
                - function (inputs: torch.Tensor, outputs: torch.Tensor, \*\*kwargs) -> Tuple[torch.Tensor, torch.Tensor] if variables_in exists. Here `inputs` is the input data and `outputs` is the output data that __getitem__ returns, so `inputs` will include the parameters if they exist. 
                - function (outputs: torch.Tensor, \*\*kwargs) -> torch.Tensor if variables_in does not exist.
                
                If batched is False, the tensors will have only one element in the first dimension.
            fn_kwargs (Dict): Additional keyword arguments to be passed to the function.
            batched (bool): If True, the function will be applied to the dataset in batches. Default is ``False``.
            batch_size (int): Size of the batch. Default is ``1000``.

        Returns:
            Dataset: A reference to the dataset with th e function applied.
        """

        if not batched:
            batch_size = 1
        pbar = tqdm(range(0, len(self), batch_size), desc="Mapping dataset", unit="iters")
        for i in pbar:
            batch = self[i:i + batch_size]
            if self.variables_in is not None:
                inputs, outputs = batch
                fn_outputs = function(inputs, outputs, **fn_kwargs)
            else:
                fn_outputs = function(batch, **fn_kwargs)
            self[i:i + batch_size] = fn_outputs

        return self

    @cr("NN.Dataset.filter")
    def filter(
        self,
        function: Callable,
        fn_kwargs: dict = {},
        batched: bool = False,
        batch_size: int = 1000,
        return_views: bool = True,
    ):
        """
        Filter the dataset using a function.

        Args:
            function (Callable): Function to be applied to the dataset with one of the following signatures:

                - function (inputs: torch.Tensor, outputs: torch.Tensor, \*\*kwargs) -> bool if variables_in exists. Here `inputs` is the input data and `outputs` is the output data that __getitem__ returns, so `inputs` will include the parameters if they exist. 
                - function (outputs: torch.Tensor, \*\*kwargs) -> bool if variables_in does not exist.

                If batched is True, the function should return a list of booleans.
            fn_kwargs (Dict): Additional keyword arguments to be passed to the function.
            batched (bool): If True, the function will be applied to the dataset in batches. Default is ``False``.
            batch_size (int): Size of the batch. Default is ``1000``.
            return_views (bool): If True, the filtered dataset will be returned as a view of the original dataset in form of torch.utils.data.Subset. Default is ``True``.
            Warning: If return_views is True, the original dataset must be kept alive, otherwise the views will point to invalid memory.
            Be careful when using `return_views=False` with `variables_in` and `parameters`, the memory usage will be higher.
            
        Returns:
            Subset | Dataset: Subset of the dataset that passed the filter.
        """

        if not batched:
            batch_size = 1
        indices = []
        pbar = tqdm(range(0, len(self), batch_size), desc="Filtering dataset", unit="iters")
        for i in pbar:
            batch = self[i:i + batch_size]
            if self.variables_in is not None:
                inputs, outputs = batch
                function_outputs = function(inputs, outputs, **fn_kwargs)
            else:
                outputs = batch
                function_outputs = function(outputs, **fn_kwargs)
            indices.extend(filter(lambda x: function_outputs[x], range(i, i + batch_size)))
        
        return self._create_subset_from_indices(indices, return_views)

    def _create_subset_from_indices(self, indices, return_views):
        if not return_views:
            data = self[indices]
            if self.variables_in is None:
                dataset = Dataset(
                    variables_out=tuple([data[:, i].reshape(-1, *self.mesh_shape).numpy()
                                for i in range(self.num_channels)]),
                    mesh_shape=self.mesh_shape,
                    variables_in=None,
                    combine_parameters_with_cartesian_prod=False,
                    snapshots_by_column=False  # Since the data is already reshaped
                )
            else:
                dataset = Dataset(
                    variables_out=tuple([data[1][:, i].reshape(-1, *self.mesh_shape).numpy()
                                for i in range(self.num_channels)]),
                    mesh_shape=self.mesh_shape,
                    variables_in=data[0].numpy(),
                    combine_parameters_with_cartesian_prod=False,
                    snapshots_by_column=False  # Since the data is already reshaped
                )
            return dataset
        else:
            return Subset(self, indices)
        
    def remove_column(self, column_idx: int, from_variables_out: bool):
        """
        Remove a column from the dataset.

        Args:
            column_idx (int): Index of the column to be removed.
            from_variables_out (bool): If True, the column will be removed from the output variables. If False, the column will be removed from the input variables, if `column_idx` is greater than the number of columns on variables_in, the column will be removed from parameters.
        """
        if from_variables_out:
            if column_idx >= self.variables_out.shape[1]:
                raiseError(f"Invalid column index {column_idx}, there are only {self.variables_out.shape[1]} columns in variables_out")
            self.variables_out = torch.cat(
                [self.variables_out[:, :column_idx], self.variables_out[:, column_idx + 1 :]], dim=1
            )
        elif column_idx < self.variables_in.shape[1]:
            self.variables_in = torch.cat(
                [self.variables_in[:, :column_idx], self.variables_in[:, column_idx + 1 :]], dim=1
            )
        elif column_idx < self.parameters.shape[1] + self.variables_in.shape[1]:
            self.parameters = torch.cat(
                [self.parameters[:, :column_idx - self.variables_in.shape[1]], self.parameters[:, column_idx - self.variables_in.shape[1] + 1 :]], dim=1
            )
        else:
            raiseError(f"Invalid column index {column_idx}, there are only {self.variables_in.shape[1]} columns in variables_in and {self.parameters.shape[1]} columns in parameters")

def set_seed(seed: int = 42, deterministic: bool = True):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
 
    if deterministic:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
 

def create_results_folder(RESUDIR: str, verbose: bool=True):
    r"""
    Create a folder to store the results of the neural network training.

    Args:
        RESUDIR (str): Path to the folder to be created.
        verbose (bool): If True, print messages to the console. Default is ``True``.
    """    
    if not os.path.exists(RESUDIR):
        os.makedirs(RESUDIR)
        if verbose: 
            print(f"Folder created: {RESUDIR}")
    elif verbose:
        print(f"Folder already exists: {RESUDIR}")


def select_device(device: str = DEVICE):
    r"""
    Select the device to be used for the training.

    Args:
        device (str): Device to be used. Default is cuda if available, otherwise cpu.
    """
    torch.device(device)
    return device


class betaLinearScheduler:
    r"""
    Linear scheduler for beta parameter in the loss function of the Autoencoders.

    Args:
        start_value (float): initial value of beta
        end_value (float): final value of beta
        warmup (int): number of epochs to reach final value
    """

    def __init__(self, start_value, end_value, start_epoch, warmup):
        self.start_value = start_value
        self.end_value = end_value
        self.start_epoch = start_epoch
        self.warmup = warmup

    def getBeta(self, epoch):
        r"""
        Get the value of beta for a given epoch.

        Args:
            epoch (int): current epoch
        """
        if epoch < self.start_epoch:
            return 0
        else:
            if epoch < self.warmup:
                beta = self.start_value + (self.end_value - self.start_value) * (
                    epoch - self.start_epoch
                ) / (self.warmup - self.start_epoch)
                return beta
            else:
                return self.end_value


#!/usr/bin/env cpython
#
# pyLOM - Python Low Order Modeling.
#
# Math operations module - regression.
#
# Last rev: 27/10/2021
from __future__ import print_function, division

import numpy as np

from ..utils.gpu import cp
from .maths      import transpose, matmul, inv
from ..utils     import cr_nvtx as cr


# This equation could be moved to the vmath module
@cr('math.least_squares')
def least_squares(A,b):
	'''
	Least squares regression
	(A^T * A)^-1 * A^T * b
	'''
	A_t = transpose(A)
	normal_matrix = matmul(A_t, A)
	inv_normal_matrix = inv(normal_matrix)
	A_t_b = matmul(A_t, b)
	return matmul(inv_normal_matrix, A_t_b)
	
@cr('math.ridge_regresion')
def ridge_regresion(A,b,lam):
	'''
	Ridge regression
	'''
	p = cp if type(A) is cp.ndarray else np
	I = p.sqrt(lam)*p.eye(A.shape[1])
	augmented_A = p.vstack([A, I])
	augmented_b = p.hstack([b,p.zeros((I.shape[0],))])
	return least_squares(augmented_A,augmented_b)


#!/usr/bin/env python
#
# Example of GPOD - Reconstruct Gappy Vector.
#
# Last revision: 26/02/2025
from __future__ import print_function, division

import numpy as np
import matplotlib.pyplot as plt
import pyLOM

pyLOM.gpu_device(gpu_per_node=4)


## Parameters
DATAFILE = "./DATA/CYLINDER.h5"
VARIABLE = "VELOX"


## Data loading
m = pyLOM.Mesh.load(DATAFILE)
d = pyLOM.Dataset.load(DATAFILE, ptable=m.partition_table).to_gpu([VARIABLE])
X = d[VARIABLE]
t = d.get_variable('time')


## Select snapshot and create gappy vector
snap = 15
np.random.seed(5)
snapshot_POD = pyLOM.GPOD.delete_snapshot(X,snap,axis=1) # Remove snapshot from training data
velox_gappy  = pyLOM.GPOD.set_random_elements_to_zero(X[:, snap], 99.9)


## Perform Gappy POD
gappy_model = pyLOM.GPOD.GappyPOD(
    centered=False,
    apply_truncation=True,
    truncation_param=-0.99,
    reconstruction_method="ridge",
    ridge_lambda=0.01,
)
gappy_model.fit(snapshot_POD)  # Fit model
velox_recons = gappy_model.predict(velox_gappy)  # Reconstruct gappy vector


## Compute and display metrics
mae_vector  = pyLOM.math.MAE(pyLOM.utils.ascontiguousarray(X[:,snap]), velox_recons)
rmse_vector = pyLOM.math.RMSE(pyLOM.utils.ascontiguousarray(X[:,snap]), velox_recons, relative=False)
r2_vector   = pyLOM.math.r2(pyLOM.utils.ascontiguousarray(X[:,snap]), velox_recons)
print(f"MAE_snapshot = {mae_vector}\nRMSE_snapshot = {rmse_vector}\nR2_snapshot = {r2_vector}")


## Add fields to dataset and send to CPU for plotting
d.add_field('VELXR',1,velox_recons)
d.add_field('ERABS',1,np.abs(X[:, snap] - velox_recons))
d.to_cpu(['VELOX','VELXR','ERABS'])

# Plot Gappy Vector
loc_gappy = np.where(velox_gappy != 0)[0]
plt.figure(figsize=(8,6),dpi=100)
plt.scatter(
    m.xyzc[loc_gappy, 0], m.xyzc[loc_gappy, 1], c=d['VELXR'][loc_gappy], cmap="jet", s=1
)
plt.title("Gappy Vector")
plt.colorbar(label="Velox")
plt.xlabel("X")
plt.ylabel("Y")

# Dump to ParaView
pyLOM.io.pv_writer(m,d,'flow',basedir='out/flow',instants=[snap],times=[t[snap]],vars=['VELOX','VELXR','ERABS'],fmt='vtkh5')
pyLOM.POD.plotSnapshot(m,d,vars=['VELXR'],instant=0,component=0,cmap='jet',cpos='xy')
pyLOM.POD.plotSnapshot(m,d,vars=['ERABS'],instant=0,component=0,cmap='coolwarm',cpos='xy')


pyLOM.cr_info()
plt.show()


#!/usr/bin/env python
#
# MDS analysis.
#
# Last revision: 11/02/2025
import matplotlib.pyplot as plt
import pyLOM

pyLOM.gpu_device(gpu_per_node=4)


## Parameters
DATAFILE = './DATA/CYLINDER.h5'
VARIABLE = 'VELOC'

## Data loading
m = pyLOM.Mesh.load(DATAFILE)
d = pyLOM.Dataset.load(DATAFILE,ptable=m.partition_table).to_gpu([VARIABLE])
X = d[VARIABLE]
t = d.get_variable('time')

Y = pyLOM.MANIFOLD.mds(X,2)

plt.scatter(Y[0,:],Y[1,:],s=10,c='b')

pyLOM.cr_info()
plt.show()



#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# Python interface for DMD.
#
# Last rev: 30/09/2021
from __future__ import print_function

import numpy as np

from ..utils.gpu import cp
from ..vmmath    import vecmat, matmul, temporal_mean, subtract_mean, tsqr_svd, transpose, eigen, cholesky, diag, polar, vandermonde, conj, inv, flip, matmulp, vandermondeTime
from ..POD       import truncate
from ..utils     import cr_nvtx as cr, cr_start, cr_stop


def _order_modes(muReal, muImag, Phi, bJov):
	'''
    Order the modes according to its amplitude, forcing that in case of a conjugate eigenvalue, the positive part always is the first one
	'''
	cnp    = cp if type(muReal) is cp.ndarray else np
	muReal = muReal[flip(cnp.abs(bJov).argsort())]
	muImag = muImag[flip(cnp.abs(bJov).argsort())]
	Phi    = transpose(transpose(Phi)[flip(cnp.abs(bJov).argsort())])
	bJov   = bJov[flip(cnp.abs(bJov).argsort())]
	p = False
	for ii in range(muImag.shape[0]):
		if p == True:
			p = False
			continue
		iimag = muImag[ii]
		if iimag < 0:
			muImag[ii]        =  muImag[ii+1]
			muImag[ii+1]      = -muImag[ii]
			bJov.imag[ii]     =  bJov.imag[ii+1]
			bJov.imag[ii+1]   = -bJov.imag[ii]
			Phi.imag[:,ii]    =  Phi.imag[:,ii+1]
			Phi.imag[:,ii+1]  = -Phi.imag[:,ii+1]
			p = True
			continue
		if iimag > 0:
			p = True
			continue
	return muReal, muImag, Phi, bJov


@cr('DMD.run')
def run(X, r, remove_mean = True):
	'''
	DMD analysis of snapshot matrix X
	Inputs:
		- X[ndims*nmesh,n_temp_snapshots]: data matrix
		- remove_mean:                     whether or not to remove the mean flow

	Returns:
		- Phi:      DMD Modes
		- muReal:   Real part of the eigenvalues
		- muImag:   Imaginary part of the eigenvalues
		- b:        Amplitude of the DMD modes
		- X_DMD:    Reconstructed flow
	'''
	# Remove temporal mean or not, depending on the user choice
	if remove_mean:
		cr_start('DMD.temporal_mean',0)
		#Compute temporal mean
		X_mean = temporal_mean(X)
		#Subtract temporal mean
		Y = subtract_mean(X, X_mean)
		cr_stop('DMD.temporal_mean',0)
	else:
		Y = X.copy()

	# Compute SVD
	cr_start('DMD.SVD',0)
	U, S, VT = tsqr_svd(Y[:, :-1])
	cr_stop('DMD.SVD',0)
	# Truncate according to residual
	cr_start('DMD.truncate', 0)
	U, S, VT = truncate(U, S, VT, r)
	cr_stop('DMD.truncate', 0)

	# Project A (Jacobian of the snapshots) into POD basis
	cr_start('DMD.linear_mapping',0)
	aux1   = matmulp(transpose(U), Y[:, 1:])
	aux2   = transpose(vecmat(1./S, VT))
	Atilde = matmul(aux1, aux2)
	cr_stop('DMD.linear_mapping',0)

	# Eigendecomposition of Atilde: Eigenvectors given as complex matrix
	# NOTE: there is no implementation of eig in cupy yet
	cr_start('DMD.modes',0)
	muReal, muImag, w = eigen(Atilde)
	muReal = cp.asarray(muReal) if type(Atilde) is cp.ndarray else muReal
	muImag = cp.asarray(muImag) if type(Atilde) is cp.ndarray else muReal
	w      = cp.asarray(w)      if type(Atilde) is cp.ndarray else muReal

	# Mode computation
	Phi =  matmul(matmul(matmul(Y[:, 1:], transpose(VT)), diag(1/S)), w)/(muReal + muImag*1J)
	cr_stop('DMD.modes',0)

	# Amplitudes according to: Jovanovic et. al. 2014 DOI: 10.1063
	cr_start('DMD.amplitudes',0)
	Vand = vandermonde(muReal, muImag, muReal.shape[0], Y.shape[1]-1)
	P    = matmul(transpose(conj(w)), w)*conj(matmul(Vand, transpose(conj(Vand))))
	Pl   = cholesky(P)
	G    = matmul(diag(S), VT)
	q    = conj(diag(matmul(matmul(Vand, transpose(conj(G))), w)))
	bJov = matmul(inv(transpose(conj(Pl))), matmul(inv(Pl), q)) #Amplitudes according to Jovanovic 2014
	cr_stop('DMD.amplitudes',0)

	# Order modes and eigenvalues according to its amplitude
	cr_start('DMD.order',0)
	muReal, muImag, Phi, bJov = _order_modes(muReal, muImag, Phi, bJov)
	cr_stop('DMD.order',0)

	return muReal, muImag, Phi, bJov

@cr('DMD.frequency_damping')
def frequency_damping(real, imag, dt):
	'''
	Computation of the damping ratio and the frequency of each mode
	'''
	p = cp if type(real) is cp.ndarray else np
	mod, arg = polar(real, imag) #Create vmmath/complex.c?
	# Computation of the damping ratio of the mode
	delta = p.log(mod)/dt
	# Computation of the frequency of the mode
	omega = arg/dt
	return delta, omega

@cr('DMD.mode_computation')
def mode_computation(X, V, S, W):
	'''
	Computation of DMD Modes
	'''
	p = cp if type(X) is cp.ndarray else np
	return  matmul(matmul(matmul(X, transpose(V)), diag(1/S)), p.abs(W))

@cr('DMD.reconstruction_jovanovic')
def reconstruction_jovanovic(Phi, real, imag, t, bJov):
	'''
    Reconstruction of the DMD modes according to the Jovanovic method
	'''
	Vand = vandermondeTime(real, imag, real.shape[0], t)
	return matmul(Phi, matmul(diag(bJov), Vand)).real



#!/usr/bin/env python
#
# PYLOM Testsuite
# Run SPOD on the cylinder dataset
#
# Last revision: 17/03/2023
from __future__ import print_function, division

import sys, os, json
import pyLOM


## Parameters
DATAFILE  = sys.argv[1]
VARIABLES = eval(sys.argv[2])
OUTDIR    = sys.argv[3]
PARAMS    = json.loads(str(sys.argv[4]).replace("'",'"'))


## Data loading
m     = pyLOM.Mesh.load(DATAFILE)
d     = pyLOM.Dataset.load(DATAFILE,ptable=m.partition_table)
X     = d.X(*VARIABLES)
t     = d.get_variable('time')
npwin = PARAMS['npwin'] #Number of snapshots in each window
nolap = PARAMS['nolap'] #Number of overlapping snapshots between windows


## Run SPOD
L, P, f = pyLOM.SPOD.run(X,t,nDFT=npwin,nolap=nolap,remove_mean=True)
if pyLOM.utils.is_rank_or_serial(root=0): 
    fig,_ = pyLOM.SPOD.plotSpectra(f, L)
    os.makedirs(OUTDIR,exist_ok=True)
    fig.savefig(f'{OUTDIR}/spectra.png',dpi=300)
pyLOM.SPOD.save(f'{OUTDIR}/results.h5',L,P,f,d.partition_table,nvars=len(VARIABLES),pointData=d.point)


## Testsuite output
M = pyLOM.SPOD.extract_modes(L,P,1,len(d),modes=[1,2,3])
pyLOM.pprint(0,'TSUITE X     =',X.min(),X.max(),X.mean())
pyLOM.pprint(0,'TSUITE L     =',L.min(),L.max(),L.mean())
pyLOM.pprint(0,'TSUITE p     =',P.min(),P.max(),P.mean())
pyLOM.pprint(0,'TSUITE f     =',f.min(),f.max(),f.mean())
pyLOM.pprint(0,'TSUITE modes =',M.min(),M.max(),M.mean())


## Dump to ParaView
# Spatial modes
d.add_field('spatial_modes',3,pyLOM.SPOD.extract_modes(L,P,1,len(d),modes=[1,2,3]))
pyLOM.io.pv_writer(m,d,'modes',basedir=f'{OUTDIR}/modes',instants=[0],times=[0.],vars=['spatial_modes'],fmt='vtkh5')


## Show and print timings
pyLOM.cr_info()
pyLOM.pprint(0,'End of output')


#!/usr/bin/env cpython
#
# pyLOM - Python Low Order Modeling.
#
# Math operations module - regression.
#
# Last rev: 27/10/2021

cimport cython
cimport numpy as np

import numpy as np

from .cfuncs   cimport real, c_sleast_squares, c_sridge_regression, c_dleast_squares, c_dridge_regression

from ..utils.cr import cr


@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.float32_t,ndim=1] _sleast_squares(float[:,:] A, float[:] b):
	'''
	Least squares regression
	(A^T * A)^-1 * A^T * b
	'''
	cdef int m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.float32_t,ndim=1] out = np.zeros((n,),dtype=np.float32)
	c_sleast_squares(&out[0],&A[0,0],&b[0],m,n)
	return out

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.double_t,ndim=1] _dleast_squares(double[:,:] A, double[:] b):
	'''
	Least squares regression
	(A^T * A)^-1 * A^T * b
	'''
	cdef int m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.double_t,ndim=1] out = np.zeros((n,),dtype=np.double)
	c_dleast_squares(&out[0],&A[0,0],&b[0],m,n)
	return out

@cr('math.least_squares')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def least_squares(real[:,:] A, real[:] b):
	'''
	Least squares regression
	(A^T * A)^-1 * A^T * b
	'''
	if real is double:
		return _dleast_squares(A,b)
	else:
		return _sleast_squares(A,b)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.float32_t,ndim=1] _sridge_regression(float[:,:] A, float[:] b, float lam):
	'''
	Least squares regression
	(A^T * A)^-1 * A^T * b
	'''
	cdef int m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.float32_t,ndim=1] out = np.zeros((n,),dtype=np.float32)
	c_sridge_regression(&out[0],&A[0,0],&b[0],lam,m,n)
	return out

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.double_t,ndim=1] _dridge_regression(double[:,:] A, double[:] b, double lam):
	'''
	Least squares regression
	(A^T * A)^-1 * A^T * b
	'''
	cdef int m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.double_t,ndim=1] out = np.zeros((n,),dtype=np.double)
	c_dridge_regression(&out[0],&A[0,0],&b[0],lam,m,n)
	return out

@cr('math.ridge_regresion')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def ridge_regresion(real[:,:] A, real[:] b, real lam):
	'''
	Ridge regression
	'''
	if real is double:
		return _dridge_regression(A,b,lam)
	else:
		return _sridge_regression(A,b,lam)